from io import StringIO

import sympy

from pharmpy import Model
from pharmpy.modeling import (
    add_peripheral_compartment,
    load_example_model,
    remove_peripheral_compartment,
    set_peripheral_compartments,
)


def create_model(s, testdata):
    model = Model(StringIO(s))
    model.dataset = load_example_model("pheno").dataset
    return model


def test_advan1(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V=VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN3 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q = QP1
V2 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.00469307) ; POP_QP1
$THETA  (0,0.011000000000000001) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct
    odes = model.statements.ode_system
    central = odes.central_compartment
    periph = odes.peripheral_compartments[0]
    rate = model.statements.ode_system.get_flow(central, periph)
    assert rate == sympy.Symbol('Q') / sympy.Symbol('V1')


def test_advan2(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN2 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V=VC
KA=1/MAT
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.00469307) ; POP_QP1
$THETA  (0,0.011000000000000001) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan2_trans1(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN2 TRANS1
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
KA=1/MAT
K=CL/VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS1
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
KA=1/MAT
K=CL/VC
K23 = QP1/VC
K32 = QP1/VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.00469307) ; POP_QP1
$THETA  (0,0.011000000000000001) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan3(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN3 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q = QP1
V2 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN11 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q2 = QP1
V2 = VP1
Q3 = QP2
V3 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.05) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.05) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan4(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN12 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q3 = QP1
V3 = VP1
Q4 = QP2
V4 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.05) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.05) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan1_two_periphs(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V=VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    add_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN11 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q2 = QP1
V2 = VP1
Q3 = QP2
V3 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.002346535) ; POP_QP1
$THETA  (0,0.011000000000000001) ; POP_VP1
$THETA  (0,0.002346535) ; POP_QP2
$THETA  (0,0.011000000000000001) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan1_remove(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V=VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    assert model.model_code == code


def test_advan3_remove(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN3 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q = QP1
V2 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V = VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,2.350801373088405) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan4_remove(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA (0,0.1) ; POP_MAT
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN2 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V = VC
KA = 1/MAT
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,2.350801373088405) ; POP_VC
$THETA (0,0.1) ; POP_MAT
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan11_remove(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN11 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q2 = QP1
V2 = VP1
Q3 = QP2
V3 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.1) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN3 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q = QP1
V2 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.2) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan12_remove(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN12 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q3 = QP1
V3 = VP1
Q4 = QP2
V4 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.1) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.2) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan11_remove_two_periphs(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN11 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q2 = QP1
V2 = VP1
Q3 = QP2
V3 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.1) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    remove_peripheral_compartment(model)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V = VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,4.48160274617681) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_advan4_roundtrip(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(5)
QP1 = THETA(4)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA (0,0.1) ; POP_MAT
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    add_peripheral_compartment(model)
    remove_peripheral_compartment(model)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN4 TRANS4
$PK
VP1 = THETA(5)
QP1 = THETA(4)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
MAT=THETA(3)*EXP(ETA(3))
V2 = VC
KA=1/MAT
Q = QP1
V3 = VP1
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA (0,0.1) ; POP_MAT
$THETA  (0,0.05) ; POP_QP1
$THETA  (0,0.2) ; POP_VP1
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$OMEGA 0.0309626  ; IVMAT
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct


def test_set_peripheral_compartments(testdata):
    code = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN11 TRANS4
$PK
VP2 = THETA(6)
QP2 = THETA(5)
VP1 = THETA(4)
QP1 = THETA(3)
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V1 = VC
Q2 = QP1
V2 = VP1
Q3 = QP2
V3 = VP2
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,0.22) ; POP_VC
$THETA  (0,0.1) ; POP_QP1
$THETA  (0,0.1) ; POP_VP1
$THETA  (0,0.1) ; POP_QP2
$THETA  (0,0.1) ; POP_VP2
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    model = create_model(code, testdata)
    set_peripheral_compartments(model, 0)
    correct = """$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV FA1 FA2
$SUBROUTINE ADVAN1 TRANS2
$PK
CL=THETA(1)*EXP(ETA(1))
VC=THETA(2)*EXP(ETA(2))
V = VC
$ERROR
Y=F+F*EPS(1)
$THETA (0,0.00469307) ; POP_CL
$THETA (0,4.48160274617681) ; POP_VC
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.0309626  ; IVV
$SIGMA 0.013241
$ESTIMATION METHOD=1 INTERACTION
"""
    assert model.model_code == correct



import numpy as np
import pandas as pd
import pytest
import sympy
from numpy.testing import assert_allclose, assert_array_equal
from pandas.testing import assert_series_equal

import pharmpy.math


def test_triangular_root():
    assert pharmpy.math.triangular_root(1) == 1
    assert pharmpy.math.triangular_root(3) == 2
    assert pharmpy.math.triangular_root(6) == 3
    assert pharmpy.math.triangular_root(10) == 4


def test_flattened_to_symmetric():
    assert_array_equal(pharmpy.math.flattened_to_symmetric([1.0]), np.array([[1.0]]))
    assert_array_equal(
        pharmpy.math.flattened_to_symmetric([1.0, 1.5, 2.0]), np.array([[1.0, 1.5], [1.5, 2.0]])
    )
    A = pharmpy.math.flattened_to_symmetric([1.0, 1.5, 2.0, -1.0, 3.0, 5.5])
    assert_array_equal(A, np.array([[1.0, 1.5, -1.0], [1.5, 2.0, 3.0], [-1.0, 3.0, 5.5]]))


def test_round_to_n_sigdig():
    assert pharmpy.math.round_to_n_sigdig(12345, 3) == 12300
    assert pharmpy.math.round_to_n_sigdig(23.45, 1) == 20
    assert pharmpy.math.round_to_n_sigdig(-0.012645, 2) == -0.013


def test_round_and_keep_sum():
    ser = pd.Series(
        [
            1.052632,
            0.701754,
            0.701754,
            1.052632,
            2.456140,
            2.807018,
            4.210526,
            4.210526,
            3.157895,
            0.350877,
        ]
    )
    correct_results = pd.Series([1, 1, 1, 1, 2, 3, 4, 4, 3, 0])
    rounded = pharmpy.math.round_and_keep_sum(ser, 20)
    assert_series_equal(rounded, correct_results)

    ser = pd.Series([1.0])
    rounded = pharmpy.math.round_and_keep_sum(ser, 1.0)
    assert_series_equal(rounded, pd.Series([1]))

    rounded = pharmpy.math.round_and_keep_sum(pd.Series([], dtype=np.float64), 1.0)
    assert_series_equal(rounded, pd.Series([], dtype=np.int64))


def test_corr2cov():
    corr = np.array([[1.00, 0.25, 0.90], [0.25, 1.00, 0.50], [0.90, 0.50, 1.00]])
    sd = np.array([1, 4, 9])
    cov = pharmpy.math.corr2cov(corr, sd)
    assert_array_equal(cov, np.array([[1, 1, 8.1], [1, 16, 18], [8.1, 18, 81]]))


def test_cov2corr():
    cov = np.array([[1.0, 1.0, 8.1], [1.0, 16.0, 18.0], [8.1, 18.0, 81.0]])
    corr = pharmpy.math.cov2corr(cov)
    assert_allclose(corr, np.array([[1.0, 0.25, 0.9], [0.25, 1, 0.5], [0.9, 0.5, 1]]))


def test_se_delta_method():
    vals = {'OMEGA(1,1)': 3.75637e-02, 'OMEGA(2,1)': 1.93936e-02, 'OMEGA(2,2)': 2.19133e-02}
    om11 = sympy.Symbol('OMEGA(1,1)')
    om21 = sympy.Symbol('OMEGA(2,1)')
    om22 = sympy.Symbol('OMEGA(2,2)')
    expr = om21 / (sympy.sqrt(om11) * sympy.sqrt(om22))
    names = ['OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)']
    cov = pd.DataFrame(
        [
            [4.17213e-04, 1.85060e-04, -3.51477e-05],
            [1.85060e-04, 1.10836e-04, 3.61663e-06],
            [-3.51477e-05, 3.61663e-06, 4.44030e-05],
        ],
        columns=names,
        index=names,
    )
    se = pharmpy.math.se_delta_method(expr, vals, cov)
    assert pytest.approx(0.2219739865800438, 1e-15) == se


def test_is_posdef():
    A = np.array([[2, -1, 0], [-1, 2, -1], [0, -1, 2]])
    assert pharmpy.math.is_posdef(A)
    B = np.array([[1, 2], [2, 1]])
    assert not pharmpy.math.is_posdef(B)


def test_nearest_posdef():
    for i in range(5):
        for j in range(2, 20):
            A = np.random.randn(j, j)
            B = pharmpy.math.nearest_posdef(A)
            assert pharmpy.math.is_posdef(B)


def test_conditional_joint_normal():
    sigma = [
        [0.0419613930249351, 0.0194493895550238, -0.00815616219453746, 0.0943578658777171],
        [0.0194493895550238, 0.0296333601234358, 0.107516199715367, 0.0353748349332184],
        [-0.00815616219453746, 0.107516199715367, 0.883267716518442, 0.101648158864576],
        [0.0943578658777171, 0.0353748349332184, 0.101648158864576, 0.887220758887173],
    ]
    sigma = np.array(sigma)

    scaling = np.diag(np.array([1, 1, 2.2376, 0.70456]))
    scaled_sigma = scaling @ sigma @ scaling.T

    WGT_mean = 1.52542372881356
    WGT_5th = 0.7
    WGT_sigma = scaled_sigma[[0, 1, 3]][:, [0, 1, 3]]
    mu = [0, 0, WGT_mean]

    mu_bar, sigma_bar = pharmpy.math.conditional_joint_normal(mu, WGT_sigma, np.array([WGT_5th]))

    np.testing.assert_array_almost_equal(mu_bar, np.array([-0.12459637, -0.04671127]))



from io import StringIO

import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, create_joint_distribution
from pharmpy.modeling.block_rvs import _choose_param_init
from pharmpy.random_variables import RandomVariable, RandomVariables
from pharmpy.results import ModelfitResults
from pharmpy.symbols import symbol as S


@pytest.mark.parametrize(
    'rvs, exception_msg',
    [
        (['ETA(3)', 'NON_EXISTENT_RV'], r'.*NON_EXISTENT_RV.*'),
        (['ETA(3)', 'ETA(6)'], r'.*ETA\(6\).*'),
        (['ETA(1)'], 'At least two random variables are needed'),
    ],
)
def test_incorrect_params(testdata, rvs, exception_msg):
    model = Model(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'multEST' / 'noSIM' / 'withBayes.mod'
    )

    with pytest.raises(Exception, match=exception_msg):
        create_joint_distribution(model, rvs)


def test_choose_param_init(pheno_path, testdata):
    model = Model(pheno_path)
    params = (model.parameters['OMEGA(1,1)'], model.parameters['OMEGA(2,2)'])
    rvs = RandomVariables(model.random_variables.etas)
    init = _choose_param_init(model, rvs, params)

    assert init == 0.0118179

    model = Model(pheno_path)
    model.modelfit_results = None
    model.name = 'run23'  # So that no results could be found
    init = _choose_param_init(model, rvs, params)

    assert init == 0.0031045

    model = Model(pheno_path)

    omega1 = S('OMEGA(3,3)')
    x = RandomVariable.normal('ETA(3)', 'IIV', 0, omega1)
    rvs.append(x)

    ie = model.modelfit_results.individual_estimates
    ie['ETA(3)'] = ie['ETA(1)']
    model.modelfit_results = ModelfitResults(individual_estimates=ie)

    init = _choose_param_init(model, rvs, params)

    assert init == 0.0118179

    # If one eta doesn't have individual estimates
    model = Model(pheno_path)
    add_iiv(model, 'S1', 'add')
    params = (model.parameters['OMEGA(1,1)'], model.parameters['IIV_S1'])
    rvs = RandomVariables([model.random_variables['ETA(1)'], model.random_variables['ETA_S1']])
    init = _choose_param_init(model, rvs, params)

    assert init == 0.0052789


def test_names(testdata):
    model = Model(
        StringIO(
            '''$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(1))
S1=V*EXP(ETA(2))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA 0.0309626  ; IVCL
$OMEGA 0.031128  ; IVV
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
        )
    )
    create_joint_distribution(model, model.random_variables.names)
    assert 'IIV_CL_V_IIV_S1' in model.parameters.names



import numpy as np
import pandas as pd

from pharmpy.modeling import copy_model, update_inits
from pharmpy.tools.modelfit import create_single_fit_workflow
from pharmpy.workflows import Task, Workflow

from .mfl import ModelFeatures


def exhaustive(base_model, mfl, run_func, rank_func):
    features = ModelFeatures(mfl)
    torun = []
    combinations = list(features.all_combinations())
    df = pd.DataFrame(
        index=pd.RangeIndex(stop=len(combinations)),
        columns=['features', 'dofv', 'rank'],
    )
    funcs = features.all_funcs()
    for i, combo in enumerate(combinations):
        model = base_model.copy()
        model.name = f'candidate{i}'
        for feat in combo:
            funcs[feat](model)
        df.loc[i]['features'] = tuple(combo)
        torun.append(model)
    run_func(torun)
    for i, model in enumerate(torun):
        df.loc[i]['dofv'] = base_model.modelfit_results.ofv - model.modelfit_results.ofv
    ranks = rank_func(base_model, torun)
    for i, ranked_model in enumerate(ranks):
        idx = torun.index(ranked_model)
        df.loc[idx]['rank'] = i + 1
    df = df.astype({'rank': 'Int64'})
    return df


def stepwise(base_model, mfl, run_func, rank_func):
    features = ModelFeatures(mfl)
    remaining = features.all_funcs()
    start_model = base_model
    current_features = []
    features_col = []
    dofv_col = []
    step = 1
    while True:
        torun = []
        for feat, func in remaining.items():
            model = start_model.copy()
            model.name = f'step_{step}_{feat.replace("(", "_").replace(")", "")}'
            func(model)
            torun.append(model)
            features_col.append(tuple(current_features + [feat]))
        print(f'Running step {step}')
        run_func(torun)
        for model in torun:
            dofv = start_model.modelfit_results.ofv - model.modelfit_results.ofv
            dofv_col.append(dofv)
        ranks = rank_func(start_model, torun)
        if not ranks:
            break
        start_model = ranks[0]
        start_model.update_inits()
        idx = torun.index(start_model)
        selected_feature = list(remaining.keys())[idx]
        current_features.append(selected_feature)
        remaining = features.next_funcs(current_features)
        step += 1
    df = pd.DataFrame({'features': features_col, 'dofv': dofv_col, 'rank': np.nan})
    best_features = tuple(current_features)
    best_df_index = features_col.index(best_features)
    df.at[best_df_index, 'rank'] = 1
    df = df.astype({'rank': 'Int64'})
    return df


def exhaustive_stepwise(mfl):
    features = ModelFeatures(mfl)
    wf_search = Workflow()

    model_tasks = []
    model_features = dict()

    candidate_count = 1

    while True:
        no_of_trans = 0

        if wf_search.output_tasks:
            actions = {
                task: _find_possible_trans(wf_search, task, features)
                for task in wf_search.output_tasks
            }
        else:
            actions = {'': (dict(), features.all_funcs())}

        for task, trans in actions.items():
            trans_previous, trans_possible = trans

            for feat, func in trans_possible.items():
                model_name = f'modelsearch_candidate{candidate_count}'

                task_copy = Task('copy', copy, model_name)
                if task:
                    wf_search.add_task(task_copy, predecessors=[task])
                else:
                    wf_search.add_task(task_copy)

                task_update_inits = Task('update_inits', update_inits)
                wf_search.add_task(task_update_inits, predecessors=task_copy)

                task_function = Task(feat, func)
                wf_search.add_task(task_function, predecessors=task_update_inits)

                wf_fit = create_single_fit_workflow()
                wf_search.insert_workflow(wf_fit, predecessors=task_function)

                model_tasks += wf_fit.output_tasks
                model_features[model_name] = tuple(list(trans_previous.keys()) + [feat])

                candidate_count += 1
                no_of_trans += 1
        if no_of_trans == 0:
            break

    return wf_search, model_tasks, model_features


def _find_possible_trans(wf, task, features):
    funcs = features.all_funcs()
    trans_previous = {
        task.name: task.function
        for task in wf.get_upstream_tasks(task)
        if task.function in funcs.values()
    }

    trans_possible = {
        feat: func
        for feat, func in funcs.items()
        if _is_allowed(feat, func, trans_previous, features)
    }

    return trans_previous, trans_possible


def _is_allowed(feat_current, func_current, trans_previous, features):
    if func_current in trans_previous.values():
        return False
    if any(func in features.get_funcs_same_type(feat_current) for func in trans_previous.values()):
        return False
    not_supported_combo = {
        'ABSORPTION(ZO)': 'TRANSITS(1)',
        'LAGTIME': 'TRANSITS',
        'TRANSITS': 'LAGTIME',
    }
    for key, value in not_supported_combo.items():
        if any(
            feat_current.startswith(key) and feat.startswith(value)
            for feat in trans_previous.keys()
        ):
            return False
    return True


def copy(name, model):
    model_copy = copy_model(model, name)
    return model_copy



from collections import OrderedDict

import pytest

from pharmpy.plugins.nonmem.records.option_record import OptionRecord


def test_create_record(parser):
    recs = parser.parse('$INPUT ID TIME DV WGT=DROP')
    rec = recs.records[0]
    pairs = rec.option_pairs
    assert pairs == OrderedDict([('ID', None), ('TIME', None), ('DV', None), ('WGT', 'DROP')])

    recs = parser.parse('$EST MAXEVAL=9999 INTERACTION')
    rec = recs.records[0]
    pairs = rec.option_pairs
    assert pairs == OrderedDict([('MAXEVAL', '9999'), ('INTERACTION', None)])


def test_all_options(parser):
    recs = parser.parse('$INPUT ID ID TIME DV WGT=DROP')
    rec = recs.records[0]
    pairs = rec.all_options
    assert pairs == [('ID', None), ('ID', None), ('TIME', None), ('DV', None), ('WGT', 'DROP')]


def test_set_option(parser):
    rec = parser.parse('$ETAS FILE=run1.phi').records[0]
    rec.set_option("FILE", "new.phi")
    assert rec.option_pairs == OrderedDict([('FILE', 'new.phi')])
    assert str(rec) == '$ETAS FILE=new.phi'

    rec = parser.parse('$EST METHOD=1 INTER ; my est\n').records[0]
    rec.set_option("METHOD", "0")
    assert str(rec) == '$EST METHOD=0 INTER ; my est\n'
    rec.set_option("CTYPE", "4")
    assert str(rec) == '$EST METHOD=0 INTER CTYPE=4 ; my est\n'


@pytest.mark.parametrize(
    "buf,remove,expected",
    [
        ('$SUBS ADVAN1 TRANS2', 'TRANS2', '$SUBS ADVAN1'),
        ('$SUBS ADVAN1  TRANS2', 'TRANS2', '$SUBS ADVAN1'),
        ('$SUBS   ADVAN1 TRANS2 ;COMMENT', 'ADVAN1', '$SUBS TRANS2 ;COMMENT'),
    ],
)
def test_remove_option(parser, buf, remove, expected):
    rec = parser.parse(buf).records[0]
    rec.remove_option(remove)
    assert str(rec) == expected


@pytest.mark.parametrize(
    "buf,remove,expected",
    [
        ('$SUBS ADVAN1 TRANS2', 'TRANS', '$SUBS ADVAN1'),
        ('$SUBS ADVAN1  TRANS2', 'TRANS', '$SUBS ADVAN1'),
        ('$SUBS   ADVAN1 TRANS2 ;COMMENT', 'ADVA', '$SUBS TRANS2 ;COMMENT'),
    ],
)
def test_remove_option_startswith(parser, buf, remove, expected):
    rec = parser.parse(buf).records[0]
    rec.remove_option_startswith(remove)
    assert str(rec) == expected


@pytest.mark.parametrize(
    "buf,expected",
    [
        ('$MODEL COMP=1 COMP=2', [['1'], ['2']]),
        ('$MODEL COMP 1 COMP 2', [['1'], ['2']]),
        ('$MODEL COMP=(CENTRAL) COMP=(PERIPHERAL)', [['CENTRAL'], ['PERIPHERAL']]),
        (
            '$MODEL COMP=(CENTRAL DEFDOSE DEFOBS) COMP=(PERIPHERAL)',
            [['CENTRAL', 'DEFDOSE', 'DEFOBS'], ['PERIPHERAL']],
        ),
        (
            '$MODEL COMP  (CENTRAL DEFDOSE DEFOBS) COMP=(PERIPHERAL)',
            [['CENTRAL', 'DEFDOSE', 'DEFOBS'], ['PERIPHERAL']],
        ),
    ],
)
def test_get_option_lists(parser, buf, expected):
    rec = parser.parse(buf).records[0]
    it = rec.get_option_lists('COMPARTMENT')
    assert list(it) == expected


@pytest.mark.parametrize(
    "buf,n,subopt,result",
    [
        (
            '$MODEL COMP=(CENTRAL) COMP=(PERIPHERAL)',
            0,
            'DEFDOSE',
            '$MODEL COMP=(CENTRAL DEFDOSE) COMP=(PERIPHERAL)',
        ),
        (
            '$MODEL COMP=(CENTRAL) COMP=(PERIPHERAL)',
            1,
            'DEFOBS',
            '$MODEL COMP=(CENTRAL) COMP=(PERIPHERAL DEFOBS)',
        ),
        (
            '$MODEL COMP=CENTRAL COMP=PERIPHERAL',
            1,
            'DEFOBS',
            '$MODEL COMP=CENTRAL COMP=(PERIPHERAL DEFOBS)',
        ),
    ],
)
def test_add_suboption_for_nth(parser, buf, n, subopt, result):
    rec = parser.parse(buf).records[0]
    rec.add_suboption_for_nth('COMPARTMENT', n, subopt)
    assert str(rec) == result


@pytest.mark.parametrize(
    "valid,opt,expected",
    [
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'DEFDOSE', 'DEFDOSE'),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'DEFDOS', 'DEFDOSE'),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'DEFD', 'DEFDOSE'),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'DEF', None),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'DEFO', 'DEFOBS'),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'NO', None),
        (['NOOFF', 'DEFDOSE', 'DEFOBS'], 'NOOF', 'NOOFF'),
    ],
)
def test_match_option(parser, valid, opt, expected):
    match = OptionRecord.match_option(valid, opt)
    assert match == expected


def test_append_node(parser):
    rec = parser.parse('$ESTIMATION METH=0 MAXEVALS=0').records[0]
    rec.append_option('INTERACTION')
    assert str(rec) == '$ESTIMATION METH=0 MAXEVALS=0 INTERACTION'
    rec.append_option('MCETA', '100')
    assert str(rec) == '$ESTIMATION METH=0 MAXEVALS=0 INTERACTION MCETA=100'


def test_prepend_node(parser):
    rec = parser.parse('$ESTIMATION METH=0 MAXEVALS=0').records[0]
    rec.prepend_option('INTERACTION')
    assert str(rec) == '$ESTIMATION INTERACTION METH=0 MAXEVALS=0'
    rec.prepend_option('MCETA', '250')
    assert str(rec) == '$ESTIMATION MCETA=250 INTERACTION METH=0 MAXEVALS=0'


def test_remove_subotion_for_all(parser):
    rec = parser.parse('$MODEL COMP=(COMP1 DEFDOSE) COMP=(COMP2)').records[0]
    rec.remove_suboption_for_all('COMPARTMENT', 'DEFDOSE')
    assert str(rec) == '$MODEL COMP=(COMP1) COMP=(COMP2)'



import re
import sys
from pathlib import Path


def options_from_command(command):
    p = re.compile('^-+([^=]+)=?(.*)')
    return {
        p.match(val).group(1): p.match(val).group(2)
        for val in command.split()
        if val.startswith('-')
    }


def arguments_from_command(command):
    command = re.sub(r'command_line:', '', command)
    return [arg for arg in command.split()[1:] if not arg.startswith('-')]


def model_paths(path, pattern, subpath='m1'):
    path = Path(path) / subpath
    model_paths = list(path.glob(pattern))
    model_paths.sort(key=lambda name: int(re.sub(r'\D', '', str(name))))
    return model_paths


def tool_from_command(command):
    command = re.sub(r'command_line:', '', command)
    tool_with_path = command.split(maxsplit=1)[0]
    return Path(tool_with_path).name


def psn_directory_list(path, drop_tools=list()):
    path = Path(path)
    folder_list = [{'name': p.name, 'tool': tool_name(p)} for p in path.iterdir() if p.is_dir()]
    return [d for d in folder_list if d['tool'] is not None and not d['tool'] in drop_tools]


def tool_name(path):
    command = psn_command(path)
    if command is not None:
        return tool_from_command(command)


def psn_command(path):
    path = Path(path)
    if not path.is_dir():
        return None
    try:
        with open(path / 'command.txt') as meta:
            return next(meta)
    except FileNotFoundError:
        pass
    try:
        with open(path / 'meta.yaml') as meta:
            cmd = None
            for row in meta:
                if row.startswith('command_line: '):
                    cmd = row.strip()
                elif cmd is not None:
                    # command can be split over multiple lines
                    if row.startswith('common_options:'):
                        return cmd
                    elif re.match(r'\s', row):  # continuation is indented
                        cmd += row.strip()
            if cmd is not None:
                return cmd
    except FileNotFoundError:
        pass


def cmd_line_model_path(path):
    path = Path(path)
    with open(path / 'meta.yaml') as meta:
        for row in meta:
            row = row.strip()
            if row.startswith('model_files:'):
                row = next(meta).strip()
                return Path(re.sub(r'^-\s*', '', row))


def template_model_string(datafile=None, ignore=list(), drop=list(), label='TEMPLATE'):
    variables = ''
    indent = r'      '
    ignores = indent + 'IGNORE=@'
    if len(ignore) > 0:
        ignores += '\n' + '\n'.join([indent + f'IGNORE=({ign})' for ign in ignore])
    if datafile is None:
        datafile = 'dummydata.csv'
    else:
        datafile = Path(datafile)
        try:
            with open(datafile) as data:
                row = next(data)
                row = row.strip()
                if re.match(r'[A-Z]', row):
                    names = row.split(r',')
                    variables = ' '
                    start = 0
                    while start < len(names):
                        stop = min(start + 10, len(names))
                        variables += (
                            ' '.join([n if n not in drop else 'DROP' for n in names[start:stop]])
                            + '\n      '
                        )
                        start = stop
                else:
                    pass  # unsupported header type
        except FileNotFoundError:
            pass
    return '\n'.join(
        [
            '$PROBLEM ' + label,
            '$INPUT' + variables,
            f'$DATA {str(datafile)}',
            ignores,
            '$SUBROUTINE ADVAN1 TRANS2',
            '',
            '$PK',
            'CL=THETA(1)*EXP(ETA(1))',
            'V=THETA(2)*EXP(ETA(2))',
            'S1=V',
            '',
            '$ERROR',
            'Y=F+F*EPS(1)',
            '',
            '$THETA (0, 1)       ; TVCL',
            '$THETA (0, 5)       ; TVV',
            '$OMEGA 0.1           ; IVCL',
            '$OMEGA 0.1           ; IVV',
            '$SIGMA 0.025         ; RUV',
            '',
            '$ESTIMATION METHOD=1 INTERACTION',
            '$COVARIANCE PRINT=E MATRIX=S UNCONDITIONAL',
        ]
    )


def pharmpy_wrapper():
    """Command line wrapper for PsN to call pharmpy"""
    args = sys.argv[1:]
    locs = dict()
    exec(args[0], globals(), locs)



import re
from math import sqrt
from pathlib import Path

import numpy as np
import pandas as pd
from scipy.linalg import cho_factor, solve_triangular

from pharmpy import Model
from pharmpy.modeling import plot_individual_predictions
from pharmpy.results import Results
from pharmpy.tools.psn_helpers import model_paths, options_from_command


class CDDResults(Results):
    """CDD Results class"""

    rst_path = Path(__file__).parent / 'report.rst'

    def __init__(self, case_results=None, case_column=None, individual_predictions_plot=None):
        self.case_results = case_results
        self.case_column = case_column
        self.individual_predictions_plot = individual_predictions_plot


def compute_cook_scores(base_estimate, cdd_estimates, covariance_matrix):
    # covariance_matrix may be jackknife estimate of covariance
    try:
        # chol * chol^T = covariance_matrix
        # inv(chol^T) * inv(chol) = inv(covariance_matrix)
        # delta_vector * inv(covariance_matrix) * delta_vector^T =
        #   delta_vector * inv(chol^T) * inv(chol)  * delta_vector^T which is
        #   the 2-norm of x   where x is
        # solution to triangular system  delta_vector^T = chol * x
        # Below we solve for all delta-vectors in one line
        chol, islow = cho_factor(covariance_matrix)
        delta_matrix = cdd_estimates - base_estimate
        x = solve_triangular(chol, delta_matrix.transpose(), lower=islow, trans=1)
        return [np.linalg.norm(v) for v in x.transpose()]
    except Exception:
        return None


def compute_delta_ofv(base_model, cdd_models, skipped_individuals):
    iofv = base_model.modelfit_results.individual_ofv
    if iofv is None:
        return [np.nan for m in cdd_models]

    cdd_ofvs = [m.modelfit_results.ofv if m.modelfit_results else np.nan for m in cdd_models]

    # need to set dtype for index.difference to work
    skipped_indices = [
        pd.Index(np.array(skipped, dtype=iofv.index.dtype)) for skipped in skipped_individuals
    ]

    return [
        sum(iofv[iofv.index.difference(skipped)]) - ofv
        for skipped, ofv in zip(skipped_indices, cdd_ofvs)
    ]


def compute_jackknife_covariance_matrix(cdd_estimates):
    bigN = len(cdd_estimates.index)
    delta_est = cdd_estimates - cdd_estimates.mean()
    return delta_est.transpose() @ delta_est * (bigN - 1) / bigN


def compute_covariance_ratios(cdd_models, covariance_matrix):
    try:
        orig_det = np.linalg.det(covariance_matrix)
        return [
            sqrt(np.linalg.det(m.modelfit_results.covariance_matrix) / orig_det)
            if m.modelfit_results and m.modelfit_results.covariance_matrix is not None
            else np.nan
            for m in cdd_models
        ]
    except Exception:
        return None


def calculate_results(base_model, cdd_models, case_column, skipped_individuals, **kwargs):
    """Calculate CDD results"""

    if base_model.modelfit_results is None:
        raise ValueError('cdd base model has no results')

    cdd_estimates = pd.DataFrame(
        data=[
            pd.Series(m.modelfit_results.parameter_estimates, name=m.name)
            for m in cdd_models
            if m.modelfit_results
        ]
    )

    cdd_model_names = [m.name for m in cdd_models]

    res = CDDResults(case_column=case_column)

    # create Series of NaN values and then replace any computable results
    cook_temp = pd.Series(np.nan, index=cdd_model_names)
    try:
        base_model.modelfit_results.covariance_matrix
    except Exception:
        pass
    else:
        cook_temp.update(
            pd.Series(
                compute_cook_scores(
                    base_model.modelfit_results.parameter_estimates,
                    cdd_estimates,
                    base_model.modelfit_results.covariance_matrix,
                ),
                index=cdd_estimates.index,
            )
        )

    jack_cook_score = None
    if len(cdd_model_names) == cdd_estimates.shape[0]:
        # all models have results
        jackkknife_covariance_matrix = compute_jackknife_covariance_matrix(cdd_estimates)
        jack_cook_score = pd.Series(
            compute_cook_scores(
                base_model.modelfit_results.parameter_estimates,
                cdd_estimates,
                jackkknife_covariance_matrix,
            ),
            index=cdd_model_names,
        )

    dofv = compute_delta_ofv(base_model, cdd_models, skipped_individuals)
    dofv_influential = [elt > 3.86 for elt in dofv]
    infl_list = [
        skipped[0]
        for skipped, infl in zip(skipped_individuals, dofv_influential)
        if infl and len(skipped) == 1
    ]

    if infl_list:
        try:
            iplot = plot_individual_predictions(
                base_model, individuals=infl_list, predictions=['PRED', 'CIPREDI']
            )
        except KeyError:
            iplot = None
    else:
        iplot = None
    res.individual_predictions_plot = iplot

    try:
        covmatrix = base_model.modelfit_results.covariance_matrix
    except Exception:
        covratios = np.nan
    else:
        covratios = compute_covariance_ratios(cdd_models, covmatrix)

    case_results = pd.DataFrame(
        {
            'cook_score': cook_temp,
            'jackknife_cook_score': jack_cook_score,
            'delta_ofv': dofv,
            'dofv_influential': dofv_influential,
            'covariance_ratio': covratios,
            'skipped_individuals': skipped_individuals,
        },
        index=cdd_model_names,
    )

    case_results.index = pd.RangeIndex(start=1, stop=len(case_results) + 1)
    res.case_results = case_results
    return res


def psn_cdd_options(path):
    path = Path(path)
    options = dict(model_path=None, outside_n_sd_check=None, case_column='ID')
    with open(path / 'meta.yaml') as meta:
        cmd = None
        for row in meta:
            if cmd is not None:
                if re.match(r'\s', row):  # continuation is indented
                    cmd += row  # must not strip
                    continue
                else:  # no continuation: parse and remove
                    for k, v in options_from_command(cmd).items():
                        if 'case_column'.startswith(k):
                            options['case_column'] = v
                            break
                    cmd = None
            row = row.strip()
            if row.startswith('model_files:'):
                row = next(meta).strip()
                options['model_path'] = re.sub(r'^-\s*', '', row)
            elif row.startswith('outside_n_sd_check: '):
                options['outside_n_sd_check'] = int(re.sub(r'\D', '', row))
            elif row.startswith('command_line: '):
                cmd = row
    return options


def psn_cdd_skipped_individuals(path):
    path = Path(path) / 'skipped_individuals1.csv'

    with open(path) as skipped:
        # rows may have different number of values, cannot use pd.read_csv
        a = [row.rstrip().split(',') for row in skipped]
    # If scientific notation convert to proper integer. Only supports integer IDs
    a = [[str(int(float(elt))) for elt in row] for row in a]
    return a


def psn_cdd_results(path, base_model_path=None):
    """Create cdd results from a PsN CDD run

    :param path: Path to PsN cdd run directory
    :return: A :class:`CDDResults` object

    """
    path = Path(path)
    if not path.is_dir():
        raise IOError(f'Could not find cdd folder: {str(path)}')

    options = psn_cdd_options(path)

    if base_model_path is None:
        base_model_path = Path(options['model_path'])
    base_model = Model(base_model_path)

    cdd_models = [Model(p) for p in model_paths(path, 'cdd_*.mod')]
    skipped_individuals = psn_cdd_skipped_individuals(path)

    res = calculate_results(base_model, cdd_models, options['case_column'], skipped_individuals)
    return res



from pathlib import Path

import numpy as np
import pandas as pd

from .common import get_model_covariates
from .data import (
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
)


def _create_dataset(model):
    idcol = model.dataset.pharmpy.id_label
    nids = get_number_of_individuals(model)
    nobs = get_number_of_observations(model)
    nobsi = get_number_of_observations_per_individual(model)
    cwres = model.modelfit_results.residuals['CWRES']

    # Max ratio of abs(ETAi) and omegai
    variance_omegas = model.random_variables.etas.variance_parameters
    omega_estimates = model.modelfit_results.parameter_estimates[variance_omegas]
    abs_ebes = model.modelfit_results.individual_estimates.abs()
    ebe_ratio = abs_ebes / list(omega_estimates)
    max_ebe_ratio = ebe_ratio.max(axis=1)

    # exp(OFVi / nobsi) / exp(OFV / nobs)
    iofv = model.modelfit_results.individual_ofv
    ofv_ratio = np.exp(iofv / nobsi) / np.exp(model.modelfit_results.ofv / nobs)

    # mean(ETA / OMEGA)
    cov = model.modelfit_results.individual_estimates_covariance
    etc_diag = pd.DataFrame([np.diag(y) for y in cov], columns=cov.iloc[0].columns)
    etc_ratio = etc_diag / list(omega_estimates)
    mean_etc_ratio = etc_ratio.mean(axis=1)
    mean_etc_ratio.index = ofv_ratio.index

    # max((abs(indcov - mean(cov))) / sd(cov))
    cov_names = get_model_covariates(model, strings=True)
    covariates = model.dataset[cov_names + [idcol]].set_index(idcol)
    mean_covs = covariates.groupby(idcol).mean()
    cov_ratio = (abs(mean_covs - mean_covs.mean()) / mean_covs.std()).mean(axis=1)

    df = pd.DataFrame(
        {
            'nids': nids,
            'nobs': nobs,
            'nobs_subj': nobsi / (nobs / nids),
            'ncovs': len(cov_names),
            'cov_ratio': cov_ratio,
            'max_cwres': abs(cwres).groupby(idcol).max(),
            'median_cwres': abs(cwres).groupby(idcol).median(),
            'max_ebe_ratio': max_ebe_ratio,
            'ofv_ratio': ofv_ratio,
            'etc_ratio': mean_etc_ratio,
        }
    )
    return df


def predict_outliers(model):
    """Predict outliers for a model using a machine learning model.

    See the :ref:`simeval <Individual OFV summary>` documentation for a definition of the `residual`

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `residual` column containing the raw predicted
        residuals and a `outlier` column with a boolean to tell whether the individual is
        an outlier or not.

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> predict_outliers(model)
        residual  outlier
    1  -0.272152    False
    2   1.723895    False
    3  -0.621257    False
    4  -0.426955    False
    5   1.639485    False
    6   1.147173    False
    7  -0.399760    False
    8   0.283858    False
    9   0.606595    False
    10 -0.505307    False
    11  5.413732     True
    12 -0.464585    False
    13  0.140510    False
    14  0.168691    False
    15 -0.472065    False
    16 -0.143271    False
    17 -0.225547    False
    18  3.730850     True
    19 -0.268493    False
    20 -0.616852    False
    21 -0.526294    False
    22 -0.393106    False
    23  1.242479    False
    24 -0.352115    False
    25  3.262130     True
    26 -0.309581    False
    27  2.062735    False
    28 -0.188629    False
    29 -0.434607    False
    30 -0.439220    False
    31 -0.227505    False
    32  3.245315     True
    33 -0.226902    False
    34  0.319238    False
    35  2.549845    False
    36 -0.086605    False
    37 -0.341915    False
    38  0.661809    False
    39 -0.678476    False
    40 -0.371514    False
    41 -0.901976    False
    42  6.411962     True
    43  1.326782    False
    44 -0.754095    False
    45 -0.453803    False
    46 -0.398793    False
    47 -0.133017    False
    48  2.460449    False
    49 -0.505857    False
    50 -0.077484    False
    51  2.011316    False
    52  1.789271    False
    53  0.579177    False
    54  0.892111    False
    55  1.943623    False
    56 -0.232907    False
    57 -0.345342    False
    58 -0.666367    False
    59 -0.578347    False

    See also
    --------
    predict_influential_individuals
    predict_influential_outliers

    """
    model_path = Path(__file__).parent.resolve() / 'ml_models' / 'outliers.tflite'
    data = _create_dataset(model)
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame({'residual': output, 'outlier': output > 3}, index=model.dataset.pharmpy.ids)
    return df


def predict_influential_individuals(model):
    """Predict influential individuals for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `dofv` column containing the raw predicted
        delta-OFV and an `influential` column with a boolean to tell whether the individual is
        influential or not.

    See also
    --------
    predict_influential_outliers
    predict_outliers

    """

    model_path = Path(__file__).parent.resolve() / 'ml_models' / 'infinds.tflite'
    data = _create_dataset(model)
    data['linearized'] = 0.0
    output = _predict_with_tflite(model_path, data)
    df = pd.DataFrame(
        {'dofv': output, 'influential': output > 3.84}, index=model.dataset.pharmpy.ids
    )
    return df


def predict_influential_outliers(model):
    """Predict influential outliers for a model using a machine learning model.

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Dataframe
        Dataframe over the individuals with a `outliers` and `dofv` columns containing the raw
        predictions and `influential`, `outlier` and `influential_outlier` boolean columns.

    See also
    --------
    predict_influential_individuals
    predict_outliers

    """

    outliers = predict_outliers(model)
    infinds = predict_influential_individuals(model)
    df = pd.concat([outliers, infinds], axis=1)
    df['influential_outlier'] = df['outlier'] & df['influential']
    return df


def _predict_with_tflite(model_path, data):
    import tflite_runtime.interpreter as tflite

    interpreter = tflite.Interpreter(str(model_path))
    interpreter.allocate_tensors()
    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()

    npdata = data.astype('float32').values
    nrows = len(data)
    output = np.empty(nrows)

    for i in range(0, nrows):
        interpreter.set_tensor(input_details[0]['index'], npdata[i : (i + 1), :])
        interpreter.invoke()
        output_data = interpreter.get_tensor(output_details[0]['index'])
        output[i] = output_data

    return output



import pharmpy.tools.modelsearch as modelsearch
import pharmpy.tools.resmod as resmod
from pharmpy.workflows import Task, Workflow


def create_workflow(model):
    wf = Workflow()
    wf.name = 'amd'

    preprocess_task = Task('preprocess', preprocess, model)

    mfl = 'ABSORPTION(ZO)\nPERIPHERALS(1)'
    wf_modelsearch = modelsearch.create_workflow('exhaustive_stepwise', mfl)
    wf.insert_workflow(wf_modelsearch, predecessors=preprocess_task)

    select_modelsearch_task = Task('select_modelsearch', select_modelsearch, model)
    wf.add_task(select_modelsearch_task, predecessors=wf_modelsearch.output_tasks)

    wf_resmod = resmod.create_workflow()
    wf.insert_workflow(wf_resmod, predecessors=wf.output_tasks)

    select_task = Task('select_resmod', select_resmod)
    wf.add_task(select_task, predecessors=wf_resmod.output_tasks)

    post_process_task = Task('results', post_process)
    wf.add_task(post_process_task, predecessors=wf.output_tasks)

    return wf


def preprocess(model):
    return model


def select_modelsearch(model, res):
    if res.best_model:
        return res.best_model
    else:
        return model


def select_resmod(res):
    return res.best_model


def post_process(*models):
    res = [model for model in models]
    return res



import os
import shutil
import time
import warnings
import weakref
from pathlib import Path
from tempfile import mkdtemp

import sympy
import sympy.physics.units as units


class TemporaryDirectoryChanger:
    def __init__(self, path):
        self.path = path
        self.old_path = Path.cwd()

    def __enter__(self):
        os.chdir(self.path)
        return self

    def __exit__(self, *args):
        os.chdir(self.old_path)


# This is copied and modified from the python 3.9 implementation
# The aim is to be able to handle Permission issues in Windows
class TemporaryDirectory:
    """Create and return a temporary directory.  This has the same
    behavior as mkdtemp but can be used as a context manager.  For
    example:

        with TemporaryDirectory() as tmpdir:
            ...

    Upon exiting the context, the directory and everything contained
    in it are removed.
    """

    def __init__(self, suffix=None, prefix=None, dir=None):
        self.name = mkdtemp(suffix, prefix, dir)
        self._finalizer = weakref.finalize(
            self, self._cleanup, self.name, warn_message="Implicitly cleaning up {!r}".format(self)
        )

    @classmethod
    def _rmtree(cls, name):
        def onerror(func, path, exc_info):
            if issubclass(exc_info[0], PermissionError):

                def resetperms(path):
                    try:
                        os.chflags(path, 0)
                    except AttributeError:
                        pass
                    os.chmod(path, 0o700)

                try:
                    if path != name:
                        resetperms(os.path.dirname(path))
                    resetperms(path)

                    try:
                        os.unlink(path)
                    # PermissionError is raised on FreeBSD for directories
                    except (IsADirectoryError, PermissionError):
                        time.sleep(0.1)
                        cls._rmtree(path)
                except FileNotFoundError:
                    pass
            elif issubclass(exc_info[0], FileNotFoundError):
                pass
            else:
                raise

        shutil.rmtree(name, onerror=onerror)

    @classmethod
    def _cleanup(cls, name, warn_message):
        cls._rmtree(name)
        warnings.warn(warn_message, ResourceWarning)

    def __repr__(self):
        return "<{} {!r}>".format(self.__class__.__name__, self.name)

    def __enter__(self):
        return self.name

    def __exit__(self, exc, value, tb):
        self.cleanup()

    def cleanup(self):
        if self._finalizer.detach():
            self._rmtree(self.name)


if os.name != 'nt':
    # Only use the custom implementation for Windows.
    from tempfile import TemporaryDirectory  # noqa


unit_subs = {}
for k, v in units.__dict__.items():
    if isinstance(v, sympy.Expr) and v.has(units.Unit):
        unit_subs[sympy.Symbol(k)] = v


def parse_units(s):
    if not isinstance(s, str):
        return s
    return sympy.sympify(s).subs(unit_subs)



import re

import pandas as pd
import pytest
from pyfakefs.fake_filesystem_unittest import Patcher

import pharmpy.plugins.nonmem as nonmem
from pharmpy import Model
from pharmpy.config import ConfigurationContext
from pharmpy.plugins.nonmem.results import NONMEMChainedModelfitResults, simfit_results


def test_ofv(pheno):
    res = pheno.modelfit_results
    assert res.ofv == 586.27605628188053


def test_aic_bic(testdata):
    model = Model(testdata / 'nonmem' / 'pheno.mod')
    assert model.modelfit_results.aic == 740.8947268137307
    assert model.modelfit_results.bic == 756.111852398327


def test_tool_files(pheno):
    res = pheno.modelfit_results
    names = [str(p.name) for p in res.tool_files]
    assert names == [
        'pheno_real.lst',
        'pheno_real.ext',
        'pheno_real.cov',
        'pheno_real.cor',
        'pheno_real.coi',
        'pheno_real.phi',
    ]


def test_condition_number(testdata, pheno):
    res = pheno.modelfit_results
    assert res.condition_number == pytest.approx(4.39152)

    maxeval3 = Model(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'oneEST' / 'noSIM' / 'maxeval3.mod'
    )

    assert maxeval3.modelfit_results.condition_number == 4.77532e06

    maxeval0 = Model(
        testdata / 'nonmem' / 'modelfit_results' / 'onePROB' / 'oneEST' / 'noSIM' / 'maxeval0.mod'
    )
    assert maxeval0.modelfit_results.condition_number is None


def test_sumo(testdata):
    onePROB = testdata / 'nonmem' / 'modelfit_results' / 'onePROB'
    pheno = Model(onePROB / 'oneEST' / 'noSIM' / 'pheno.mod')
    from pharmpy.modeling import load_example_model

    pheno = load_example_model("pheno")
    d = pheno.modelfit_results.sumo(to_string=False)
    assert 'Messages' in d.keys()
    assert 'Parameter summary' in d.keys()


def test_special_models(testdata):
    onePROB = testdata / 'nonmem' / 'modelfit_results' / 'onePROB'
    withBayes = Model(onePROB / 'multEST' / 'noSIM' / 'withBayes.mod')
    assert (
        pytest.approx(withBayes.modelfit_results.standard_errors['THETA(1)'], 1e-13) == 2.51942e00
    )
    assert (
        pytest.approx(withBayes.modelfit_results[0].standard_errors['THETA(1)'], 1e-13)
        == 3.76048e-01
    )
    assert withBayes.modelfit_results[0].minimization_successful is False
    assert withBayes.modelfit_results[1].minimization_successful is False
    assert withBayes.modelfit_results[0].covariance_step == {
        'requested': True,
        'completed': True,
        'warnings': False,
    }
    assert withBayes.modelfit_results.covariance_step == {
        'requested': True,
        'completed': True,
        'warnings': False,
    }

    maxeval0 = Model(onePROB / 'oneEST' / 'noSIM' / 'maxeval0.mod')
    assert maxeval0.modelfit_results.minimization_successful is None

    maxeval3 = Model(onePROB / 'oneEST' / 'noSIM' / 'maxeval3.mod')
    assert maxeval3.modelfit_results.minimization_successful is False
    assert maxeval3.modelfit_results.covariance_step == {
        'requested': True,
        'completed': True,
        'warnings': True,
    }

    nearbound = Model(onePROB / 'oneEST' / 'noSIM' / 'near_bounds.mod')
    correct = pd.Series(
        [False, True, False, False, False, False, False, False, True, True, False],
        index=[
            'THETA(1)',
            'THETA(2)',
            'THETA(3)',
            'THETA(4)',
            'OMEGA(1,1)',
            'OMEGA(2,1)',
            'OMEGA(2,2)',
            'OMEGA(3,3)',
            'OMEGA(4,4)',
            'OMEGA(6,6)',
            'SIGMA(1,1)',
        ],
    )
    pd.testing.assert_series_equal(nearbound.modelfit_results.near_bounds(), correct)


def test_covariance(pheno_path):
    with ConfigurationContext(nonmem.conf, parameter_names=['basic']):
        res = Model(pheno_path).modelfit_results
        cov = res.covariance_matrix
        assert len(cov) == 6
        assert pytest.approx(cov.loc['THETA(1)', 'THETA(1)'], 1e-13) == 4.41151e-08
        assert pytest.approx(cov.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == 7.17184e-05
    with ConfigurationContext(nonmem.conf, parameter_names=['comment', 'basic']):
        res = Model(pheno_path).modelfit_results
        cov = res.covariance_matrix
        assert len(cov) == 6
        assert pytest.approx(cov.loc['PTVCL', 'PTVCL'], 1e-13) == 4.41151e-08
        assert pytest.approx(cov.loc['IVV', 'PTVV'], 1e-13) == 7.17184e-05


def test_information(pheno_path):
    with ConfigurationContext(nonmem.conf, parameter_names=['basic']):
        res = Model(pheno_path).modelfit_results
        m = res.information_matrix
        assert len(m) == 6
        assert pytest.approx(m.loc['THETA(1)', 'THETA(1)'], 1e-13) == 2.99556e07
        assert pytest.approx(m.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == -2.80082e03
    with ConfigurationContext(nonmem.conf, parameter_names=['comment', 'basic']):
        res = Model(pheno_path).modelfit_results
        m = res.information_matrix
        assert len(m) == 6
        assert pytest.approx(m.loc['PTVCL', 'PTVCL'], 1e-13) == 2.99556e07
        assert pytest.approx(m.loc['IVV', 'PTVV'], 1e-13) == -2.80082e03


def test_correlation(pheno_path):
    with ConfigurationContext(nonmem.conf, parameter_names=['basic']):
        res = Model(pheno_path).modelfit_results
        corr = res.correlation_matrix
        assert len(corr) == 6
        assert corr.loc['THETA(1)', 'THETA(1)'] == 1.0
        assert pytest.approx(corr.loc['OMEGA(2,2)', 'THETA(2)'], 1e-13) == 3.56662e-01
    with ConfigurationContext(nonmem.conf, parameter_names=['comment', 'basic']):
        res = Model(pheno_path).modelfit_results
        corr = res.correlation_matrix
        assert len(corr) == 6
        assert corr.loc['PTVCL', 'PTVV'] == 0.00709865
        assert pytest.approx(corr.loc['IVV', 'PTVV'], 1e-13) == 3.56662e-01


def test_standard_errors(testdata, pheno_path):
    with ConfigurationContext(nonmem.conf, parameter_names=['basic']):
        res = Model(pheno_path).modelfit_results
        ses = res.standard_errors
        assert len(ses) == 6
        assert pytest.approx(ses['THETA(1)'], 1e-13) == 2.10036e-04
        ses_sd = res.standard_errors_sdcorr
        correct = pd.Series(
            {
                'THETA(1)': 0.000210036,
                'THETA(2)': 0.0268952,
                'THETA(3)': 0.0837623,
                'OMEGA(1,1)': 0.0391526,
                'OMEGA(2,2)': 0.0223779,
                'SIGMA(1,1)': 0.00990444,
            }
        )
        correct.name = 'SE'
        pd.testing.assert_series_equal(ses_sd, correct)

    with ConfigurationContext(nonmem.conf, parameter_names=['comment', 'basic']):
        res = Model(pheno_path).modelfit_results
        ses = res.standard_errors
        assert len(ses) == 6
        assert pytest.approx(ses['PTVCL'], 1e-13) == 2.10036e-04

        ses_sd = res.standard_errors_sdcorr
        correct = pd.Series(
            {
                'PTVCL': 0.000210036,
                'PTVV': 0.0268952,
                'THETA(3)': 0.0837623,
                'IVCL': 0.0391526,
                'IVV': 0.0223779,
                'SIGMA(1,1)': 0.00990444,
            }
        )
        correct.name = 'SE'
        pd.testing.assert_series_equal(ses_sd, correct)


def test_individual_ofv(pheno):
    iofv = pheno.modelfit_results.individual_ofv
    assert len(iofv) == 59
    assert pytest.approx(iofv[1], 1e-15) == 5.9473520242962552
    assert pytest.approx(iofv[57], 1e-15) == 5.6639479151436394


def test_individual_estimates(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, model=pheno)
    ie = res.individual_estimates
    assert len(ie) == 59
    assert pytest.approx(ie['ETA(1)'][1], 1e-15) == -0.0438608
    assert pytest.approx(ie['ETA(2)'][1], 1e-15) == 0.00543031
    assert pytest.approx(ie['ETA(1)'][28], 1e-15) == 7.75957e-04
    assert pytest.approx(ie['ETA(2)'][28], 1e-15) == 8.32311e-02


def test_individual_estimates_covariance(pheno, pheno_lst):
    res = NONMEMChainedModelfitResults(pheno_lst, model=pheno)
    cov = res.individual_estimates_covariance
    assert len(cov) == 59
    names = ['ETA(1)', 'ETA(2)']
    correct = pd.DataFrame(
        [[2.48833e-02, -2.99920e-03], [-2.99920e-03, 7.15713e-03]], index=names, columns=names
    )
    pd.testing.assert_frame_equal(cov[1], correct)
    correct2 = pd.DataFrame(
        [[2.93487e-02, -1.95747e-04], [-1.95747e-04, 8.94118e-03]], index=names, columns=names
    )
    pd.testing.assert_frame_equal(cov[43], correct2)


def test_parameter_estimates(pheno_path):
    with ConfigurationContext(nonmem.conf, parameter_names=['basic']):
        res = Model(pheno_path).modelfit_results
        pe = res.parameter_estimates
        assert len(pe) == 6
        assert pe['THETA(1)'] == 4.69555e-3
        assert pe['OMEGA(2,2)'] == 2.7906e-2
        pe_sd = res.parameter_estimates_sdcorr
        correct = pd.Series(
            {
                'THETA(1)': 0.00469555,
                'THETA(2)': 0.984258,
                'THETA(3)': 0.158920,
                'OMEGA(1,1)': 0.171321,
                'OMEGA(2,2)': 0.167051,
                'SIGMA(1,1)': 0.115069,
            }
        )
        correct.name = 'estimates'
        pd.testing.assert_series_equal(pe_sd, correct)

    with ConfigurationContext(nonmem.conf, parameter_names=['comment', 'basic']):
        res = Model(pheno_path).modelfit_results
        pe = res.parameter_estimates
        assert len(pe) == 6
        assert pe['PTVCL'] == 4.69555e-3
        assert pe['IVV'] == 2.7906e-2


def test_simfit(testdata):
    model = Model(testdata / 'nonmem' / 'modelfit_results' / 'simfit' / 'sim-1.mod')
    results = simfit_results(model)
    assert len(results) == 3
    assert results[1].ofv == 565.84904364342981
    assert results[2].ofv == 570.73440114145342


def test_residuals(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    df = model.modelfit_results.residuals
    assert len(df) == 155
    assert list(df.columns) == ['RES', 'CWRES']
    assert df['RES'][1.0, 2.0] == -0.67071
    assert df['CWRES'][1.0, 2.0] == -0.401100


def test_predictions(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    df = model.modelfit_results.predictions
    assert len(df) == 744
    assert list(df.columns) == ['IPRED', 'PRED']
    assert df['PRED'][1.0, 0.0] == 18.143


def test_runtime_total(testdata):
    model = Model(testdata / 'nonmem' / 'pheno_real.mod')
    runtime = model.modelfit_results.runtime_total
    assert runtime == 4


@pytest.mark.parametrize(
    'starttime, endtime, runtime_ref',
    [
        (
            'Sat Sep 8 10:57:25 CEST 2018',
            'Sat Sep 8 10:57:29 CEST 2018',
            4,
        ),
        (
            '08/09/2018\n10:57:25',
            '08/09/2018\n10:57:29',
            4,
        ),
        (
            '2018-09-08\n10:57:25',
            '2018-09-08\n10:57:29',
            4,
        ),
        (
            '2018-09-08\n10:57',
            '2018-09-08\n10:57',
            0,
        ),
        (
            '2018-09-08\n10:57',
            '2018-09-08\n10:58',
            60,
        ),
    ],
)
def test_runtime_different_formats(testdata, starttime, endtime, runtime_ref):
    with open(testdata / 'nonmem' / 'pheno_real.lst', encoding='utf-8') as lst_file:
        lst_file_str = lst_file.read()

        repl_dict = {
            'start': ('lör  8 sep 2018 10:57:25 CEST', starttime),
            'end': ('lör  8 sep 2018 10:57:29 CEST', endtime),
        }

        lst_file_repl = re.sub(
            repl_dict['start'][0],
            repl_dict['start'][1],
            lst_file_str,
        )
        lst_file_repl = re.sub(
            repl_dict['end'][0],
            repl_dict['end'][1],
            lst_file_repl,
        )
        assert repl_dict['start'][0] not in lst_file_repl
        assert repl_dict['end'][0] not in lst_file_repl

    with Patcher(additional_skip_names=['pkgutil']) as patcher:
        fs = patcher.fs
        fs.add_real_file(testdata / 'nonmem' / 'pheno_real.mod', target_path='pheno_real.mod')
        fs.add_real_file(testdata / 'nonmem' / 'pheno_real.ext', target_path='pheno_real.ext')

        with open('pheno_real.lst', 'a') as f:
            f.write(lst_file_repl)

        model = Model('pheno_real.mod')
        runtime = model.modelfit_results.runtime_total
        assert runtime == runtime_ref


def test_result_summary(pheno_path, testdata):
    model = Model(pheno_path)
    res = model.modelfit_results

    summary = res.result_summary()

    assert summary.loc['pheno_real']['ofv'] == 586.2760562818805
    assert summary.loc['pheno_real']['OMEGA(1,1)_estimate'] == 0.0293508

    model = Model(
        testdata
        / 'nonmem'
        / 'modelfit_results'
        / 'onePROB'
        / 'multEST'
        / 'noSIM'
        / 'pheno_multEST.mod'
    )
    res = model.modelfit_results
    summary = res.result_summary(include_all_estimation_steps=True)

    assert not summary.loc['pheno_multEST', 1]['minimization_successful']
    assert summary.loc['pheno_multEST', 2]['minimization_successful']
    assert summary.loc['pheno_multEST', 1]['run_type'] == 'estimation'
    assert summary.loc['pheno_multEST', 2]['run_type'] == 'evaluation'

    assert summary.loc['pheno_multEST', 1]['ofv'] != summary.loc['pheno_multEST', 2]['ofv']


def test_evaluation(testdata):
    model = Model(
        testdata
        / 'nonmem'
        / 'modelfit_results'
        / 'onePROB'
        / 'multEST'
        / 'noSIM'
        / 'pheno_multEST.mod'
    )
    res = model.modelfit_results

    assert round(res.ofv, 3) == 729.955
    assert res[-1].minimization_successful
    assert not res.minimization_successful



class Record:
    """
    Top level class for records.

    Create objects only by using the factory function create_record.
    """

    name = None

    def __init__(self, content, parser_class):
        self._content = content
        self._parser_class = parser_class
        self._root = None

    @property
    def root(self):
        """Root of the parse tree
        only parse on demand
        """
        if self._root is None:
            parser = self._parser_class(self._content)
            self._root = parser.root
            del self._parser_class
            del self._content
        return self._root

    @root.setter
    def root(self, root_new):
        self._root = root_new

    def __str__(self):
        if self._root is None:
            return self.raw_name + self._content
        else:
            return self.raw_name + str(self.root)



import copy
import pickle

import pytest
import sympy
import sympy.stats as stats

from pharmpy.random_variables import RandomVariable, RandomVariables, VariabilityHierarchy
from pharmpy.symbols import symbol


def test_general_rv():
    rv = RandomVariable('X', 'iiv', stats.Normal('X', 0, 1))
    assert rv.name == 'X'
    assert rv.symbol == symbol('X')
    assert rv.level == 'IIV'
    assert rv.sympy_rv.pspace.distribution.mean == 0
    with pytest.raises(ValueError):
        RandomVariable('X', 'iiv', stats.Normal('X', [0, 0], [[1, 0], [0, 2]]))
    rv = RandomVariable('X', 'iiv', stats.Exponential('X', 1))


def test_normal_rv():
    rv = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    assert rv.name == 'ETA(1)'
    assert rv.symbol == symbol('ETA(1)')
    assert rv.level == 'IIV'
    assert rv.joint_names == []
    rv.name = 'NEW'
    assert rv.name == 'NEW'
    with pytest.raises(ValueError):
        RandomVariable.normal('ETA(1)', 'uuu', 0, 1)
    with pytest.raises(ValueError):
        RandomVariable.normal('X', 'iiv', 0, -1)


def test_joint_normal_rv():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    assert rv1.name == 'ETA(1)'
    assert rv2.name == 'ETA(2)'
    assert rv1.symbol == symbol('ETA(1)')
    assert rv2.symbol == symbol('ETA(2)')
    assert rv1.level == 'IIV'
    assert rv2.level == 'IIV'
    assert rv1.joint_names == ['ETA(1)', 'ETA(2)']
    assert rv2.joint_names == ['ETA(1)', 'ETA(2)']
    rv1.name = 'NEW'
    assert rv1.name == 'NEW'
    with pytest.raises(ValueError):
        RandomVariable.joint_normal(['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[-1, 0.1], [0.1, 2]])


def test_eq_rv():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    assert rv1 == rv2
    rv3 = RandomVariable.normal('ETA(2)', 'iiv', 0, 1)
    assert rv1 != rv3
    rv4 = RandomVariable.normal('ETA(2)', 'iiv', 0, 0.1)
    assert rv3 != rv4


def test_sympy_rv():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    assert rv1.sympy_rv == sympy.stats.Normal('ETA(1)', 0, 1)
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 0)
    assert rv1.sympy_rv == sympy.Integer(0)


def test_repr_rv():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    assert repr(rv1) == 'ETA(1) ~ 𝒩 (0, 1)'
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    assert (
        repr(rv1)
        == """⎡ETA(1)⎤     ⎧⎡0⎤  ⎡ 1   0.1⎤⎫
⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦     ⎩⎣0⎦  ⎣0.1   2 ⎦⎭"""
    )


def test_repr_latex_rv():
    rv1, rv2 = RandomVariable.joint_normal(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    assert (
        rv1._repr_latex_()
        == '$\\displaystyle \\left[\\begin{matrix}x\\\\y\\end{matrix}\\right]\\sim \\mathcal{N} \\left(\\displaystyle \\left[\\begin{matrix}0\\\\0\\end{matrix}\\right],\\displaystyle \\left[\\begin{matrix}1 & 0.1\\\\0.1 & 2\\end{matrix}\\right]\\right)$'  # noqa E501
    )

    rv = RandomVariable.normal('x', 'iiv', 0, 1)
    assert (
        rv._repr_latex_()
        == '$\\displaystyle x\\sim  \\mathcal{N} \\left(\\displaystyle 0,\\displaystyle 1\\right)$'
    )


def test_copy_rv():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = rv1.copy()
    assert rv1 is not rv2
    assert rv1 == rv2


def test_parameters_rv():
    rv1 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    assert rv1.parameter_names == ['OMEGA(2,2)']


def test_init_from_rvs():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 1)
    rvs = RandomVariables([rv1, rv2])
    rvs2 = RandomVariables(rvs)
    assert len(rvs2) == 2


def test_illegal_inits():
    with pytest.raises(ValueError):
        RandomVariables([8, 1])
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    with pytest.raises(ValueError):
        RandomVariables([rv1, rv2])


def test_len():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 1)
    rvs = RandomVariables([rv1, rv2])
    assert len(rvs) == 2


def test_eq():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 1)
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, 1)
    rvs = RandomVariables([rv1, rv2])
    rvs2 = RandomVariables([rv1])
    assert rvs != rvs2
    rvs3 = RandomVariables([rv1, rv3])
    assert rvs != rvs3


def test_getitem():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2])
    assert rvs[0] == rv1
    assert rvs[1] == rv2
    assert rvs['ETA(1)'] == rv1
    assert rvs['ETA(2)'] == rv2
    assert rvs[symbol('ETA(1)')] == rv1
    assert rvs[symbol('ETA(2)')] == rv2
    with pytest.raises(IndexError):
        rvs[23]
    with pytest.raises(KeyError):
        rvs['NOKEYOFTHIS']

    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    selection = rvs[[rv1, rv2]]
    assert len(selection) == 2


def test_contains():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    assert 'ETA(2)' in rvs
    assert 'ETA(4)' not in rvs


def test_setitem():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1])
    rvs[rv1] = rv2
    assert len(rvs) == 1
    assert rvs[0].name == 'ETA(2)'
    with pytest.raises(KeyError):
        rvs[rv3] = rv2

    rv1, rv2 = RandomVariable.joint_normal(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    rv3 = RandomVariable.normal('z', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2])
    rvs['x'] = rv3

    rv1, rv2, rv3 = RandomVariable.joint_normal(
        ['x', 'y', 'z'], 'iiv', [0, 0, 1], [[1, 0.1, 0.1], [0.1, 4, 0.1], [0.1, 0.1, 9]]
    )
    rv4 = RandomVariable.normal('w', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs['y'] = rv4
    cov = sympy.Matrix([[1, 0.1], [0.1, 9]])
    assert rv1.sympy_rv.pspace.distribution.sigma == cov
    assert rv3.sympy_rv.pspace.distribution.sigma == cov
    assert rvs.names == ['w', 'x', 'z']

    with pytest.raises(ValueError):
        rvs[0] = 0

    rv1, rv2, rv3 = RandomVariable.joint_normal(
        ['x', 'y', 'z'], 'iiv', [0, 0, 1], [[1, 0.1, 0.1], [0.1, 4, 0.1], [0.1, 0.1, 9]]
    )
    rv4 = RandomVariable.normal('w', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs[0:1] = [rv4]
    assert len(rvs) == 3
    assert rvs.names == ['w', 'y', 'z']

    with pytest.raises(ValueError):
        rvs[0:2] = [rv4]

    rv1, rv2, rv3 = RandomVariable.joint_normal(
        ['x', 'y', 'z'], 'iiv', [0, 0, 1], [[1, 0.1, 0.1], [0.1, 4, 0.1], [0.1, 0.1, 9]]
    )
    rv4 = RandomVariable.normal('w', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs[0:3:2] = [rv1, rv4]


def test_insert():
    rv1, rv2 = RandomVariable.joint_normal(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    rvs = RandomVariables([rv1, rv2])
    with pytest.raises(ValueError):
        rvs.insert(0, 23)


def test_delitem():
    rv1 = RandomVariable.normal('ETA', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA2', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    del rvs['ETA']
    assert len(rvs) == 2
    assert rvs.names == ['ETA2', 'EPS']

    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rvs = RandomVariables([rv1, rv2])
    del rvs[0]
    assert len(rvs) == 1

    rv1, rv2, rv3 = RandomVariable.joint_normal(
        ['x', 'y', 'z'], 'iiv', [0, 0, 1], [[1, 0.1, 0.1], [0.1, 4, 0.1], [0.1, 0.1, 9]]
    )
    rvs = RandomVariables([rv1, rv2, rv3])
    del rvs[1]
    assert len(rvs) == 2


def test_names():
    rv1 = RandomVariable.normal('ETA1', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA2', 'iiv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2])
    assert rvs.names == ['ETA1', 'ETA2']


def test_epsilons():
    rv1 = RandomVariable.normal('ETA', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA2', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.epsilons == RandomVariables([rv3])
    assert rvs.epsilons.names == ['EPS']


def test_etas():
    rv1 = RandomVariable.normal('ETA', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA2', 'iiv', 0, 0.1)
    rv3 = RandomVariable.normal('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.etas == RandomVariables([rv1, rv2])
    assert rvs.etas.names == ['ETA', 'ETA2']


def test_iiv_iov():
    rv1 = RandomVariable.normal('ETA', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA2', 'iov', 0, 0.1)
    rv3 = RandomVariable.normal('EPS', 'ruv', 0, 0.1)
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.iiv == RandomVariables([rv1])
    assert rvs.iiv.names == ['ETA']
    assert rvs.iov == RandomVariables([rv2])
    assert rvs.iov.names == ['ETA2']


def test_subs():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs.subs(
        {
            symbol('ETA(2)'): symbol('w'),
            symbol('OMEGA(1,1)'): symbol('x'),
            symbol('OMEGA(3,3)'): symbol('y'),
        }
    )
    assert rv1.sympy_rv.pspace.distribution.sigma == sympy.Matrix(
        [[symbol('x'), symbol('OMEGA(2,1)')], [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')]]
    )
    assert rv3.sympy_rv.pspace.distribution.std ** 2 == symbol('y')
    assert rvs.names == ['ETA(1)', 'w', 'ETA(3)']


def test_free_symbols():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    assert rv2.free_symbols == {symbol('ETA(2)'), symbol('OMEGA(2,2)')}
    rvs = RandomVariables([rv1, rv2])
    assert rvs.free_symbols == {symbol('ETA(1)'), symbol('ETA(2)'), symbol('OMEGA(2,2)')}
    rv_exp = RandomVariable('X', 'iiv', stats.Exponential('X', 1))
    assert rv_exp.free_symbols == {symbol('X')}


def test_parameter_names():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    assert rv1.parameter_names == ['OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)']
    assert rv2.parameter_names == ['OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)']
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    assert rv3.parameter_names == ['OMEGA(3,3)']
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.parameter_names == ['OMEGA(1,1)', 'OMEGA(2,1)', 'OMEGA(2,2)', 'OMEGA(3,3)']
    rv_exp = RandomVariable('X', 'iiv', stats.Exponential('X', symbol('Z')))
    assert rv_exp.parameter_names == ['Z']


def test_subs_rv():
    rv = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rv.subs({symbol('OMEGA(3,3)'): symbol('VAR')})
    assert rv.sympy_rv.pspace.distribution.std == sympy.sqrt(symbol('VAR'))
    rv_exp = RandomVariable('X', 'iiv', stats.Exponential('X', symbol('Z')))
    rv_exp.subs({symbol('Z'): symbol('TV')})
    assert rv_exp.sympy_rv.pspace.distribution.rate == symbol('TV')


def test_distributions():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rvs = RandomVariables([rv1, rv2])
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0.5, 2)
    rvs.append(rv3)
    dists = rvs.distributions()
    symbols, dist = dists[0]
    assert symbols[0].name == 'ETA(1)'
    assert symbols[1].name == 'ETA(2)'
    assert len(symbols) == 2
    assert dist == rv1.sympy_rv.pspace.distribution
    symbols, dist = dists[1]
    assert symbols[0].name == 'ETA(3)'
    assert len(symbols) == 1
    assert dist == rv3.sympy_rv.pspace.distribution


def test_repr():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables([rv1, rv2, rv3])
    res = """⎡ETA(1)⎤     ⎧⎡0⎤  ⎡ 1   0.1⎤⎫
⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦     ⎩⎣0⎦  ⎣0.1   2 ⎦⎭
ETA(3) ~ 𝒩 (2, 1)"""
    assert str(rvs) == res
    rv_exp = RandomVariable('X', 'iiv', stats.Exponential('X', symbol('Z')))
    assert str(rv_exp) == 'X ~ Exp(Z)'
    rv_f = RandomVariable('X', 'iiv', stats.FDistribution('X', symbol('Z'), 2))
    assert str(rv_f) == 'X ~ UnknownDistribution'
    rv3, rv4 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [sympy.sqrt(sympy.Rational(2, 5)), 0], [[1, 0.1], [0.1, 2]]
    )
    assert (
        str(rv3)
        == '''             ⎧⎡√10⎤            ⎫
⎡ETA(1)⎤     ⎪⎢───⎥  ⎡ 1   0.1⎤⎪
⎢      ⎥ ~ 𝒩 ⎪⎢ 5 ⎥, ⎢        ⎥⎪
⎣ETA(2)⎦     ⎪⎢   ⎥  ⎣0.1   2 ⎦⎪
             ⎩⎣ 0 ⎦            ⎭'''
    )


def test_repr_latex():
    rv1 = RandomVariable.normal('z', 'iiv', 0, 1)
    rv2, rv3 = RandomVariable.joint_normal(['x', 'y'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]])
    rvs = RandomVariables([rv1, rv2, rv3])
    assert (
        rvs._repr_latex_()
        == '\\begin{align*}\n\\displaystyle z & \\sim  \\mathcal{N} \\left(\\displaystyle 0,\\displaystyle 1\\right) \\\\ \\displaystyle \\left[\\begin{matrix}x\\\\y\\end{matrix}\\right] & \\sim \\mathcal{N} \\left(\\displaystyle \\left[\\begin{matrix}0\\\\0\\end{matrix}\\right],\\displaystyle \\left[\\begin{matrix}1 & 0.1\\\\0.1 & 2\\end{matrix}\\right]\\right)\\end{align*}'  # noqa E501
    )


def test_copy():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs2 = rvs.copy()
    assert rvs == rvs2
    assert rvs is not rvs2
    rvs3 = copy.copy(rvs2)
    assert rvs == rvs3
    assert rvs is not rvs3


def test_pickle():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'], 'iiv', [0, 0], [[1, 0.1], [0.1, 2]]
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 2, 1)
    rvs = RandomVariables([rv1, rv2, rv3])
    pickled = pickle.dumps(rvs)
    obj = pickle.loads(pickled)
    assert obj == rvs


def test_hash():
    rv1 = RandomVariable.normal('ETA(3)', 'iiv', 2, 1)
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 2, 0)
    assert hash(rv1) != hash(rv2)


def test_nearest_valid_parameters():
    values = {'x': 1, 'y': 0.1, 'z': 2}
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('x'), symbol('y')], [symbol('y'), symbol('z')]],
    )
    rvs = RandomVariables([rv1, rv2])
    new = rvs.nearest_valid_parameters(values)
    assert values == new
    values = {'x': 1, 'y': 1.1, 'z': 1}
    new = rvs.nearest_valid_parameters(values)
    assert new == {'x': 1.0500000000000005, 'y': 1.0500000000000003, 'z': 1.050000000000001}


def test_validate_parameters():
    a, b, c, d = (symbol('a'), symbol('b'), symbol('c'), symbol('d'))
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[a, b], [c, d]],
    )
    rvs = RandomVariables([rv1, rv2])
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0.5, d)
    rvs.append(rv3)
    params = {'a': 2, 'b': 0.1, 'c': 1, 'd': 23}
    assert rvs.validate_parameters(params)
    params2 = {'a': 2, 'b': 2, 'c': 23, 'd': 1}
    assert not rvs.validate_parameters(params2)
    with pytest.raises(TypeError):
        rvs.validate_parameters({})


def test_sample():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('a'), symbol('b')], [symbol('b'), symbol('c')]],
    )
    rvs = RandomVariables([rv1, rv2])
    params = {'a': 1, 'b': 0.1, 'c': 2}
    samples = rvs.sample(rv1.symbol + rv2.symbol, parameters=params, samples=2, rng=9532)
    assert list(samples) == pytest.approx([1.7033555824617346, -1.4031809274765599])
    with pytest.raises(TypeError):
        rvs.sample(rv1.symbol + rv2.symbol, samples=1, rng=9532)
    samples = rvs.sample(1, samples=2)
    assert list(samples) == [1.0, 1.0]


def test_variance_parameters():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.variance_parameters == ['OMEGA(1,1)', 'OMEGA(2,2)', 'OMEGA(3,3)']

    rv1 = RandomVariable.normal('x', 'iiv', 0, symbol('omega'))
    rv2 = RandomVariable.normal('y', 'iiv', 0, symbol('omega'))
    rvs = RandomVariables([rv1, rv2])
    assert rvs.variance_parameters == ['omega']


def test_get_variance():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.get_variance('ETA(1)') == symbol('OMEGA(1,1)')
    assert rvs.get_variance(rv3) == symbol('OMEGA(3,3)')


def test_get_covariance():
    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(1,1)'), symbol('OMEGA(2,1)')],
            [symbol('OMEGA(2,1)'), symbol('OMEGA(2,2)')],
        ],
    )
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2, rv3])
    assert rvs.get_covariance('ETA(1)', 'ETA(2)') == symbol('OMEGA(2,1)')
    assert rvs.get_covariance(rv3, rv2) == 0


def test_join():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables([rv1, rv2])
    rvs.join(['ETA(1)', 'ETA(2)'])
    assert rv1.sympy_rv.pspace.distribution.sigma == sympy.Matrix(
        [[symbol('OMEGA(1,1)'), 0], [0, symbol('OMEGA(2,2)')]]
    )
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables([rv1, rv2])
    rvs.join(['ETA(1)', 'ETA(2)'], fill=1)
    assert rv1.sympy_rv.pspace.distribution.sigma == sympy.Matrix(
        [[symbol('OMEGA(1,1)'), 1], [1, symbol('OMEGA(2,2)')]]
    )
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables([rv1, rv2])
    rvs.join(['ETA(1)', 'ETA(2)'], name_template='IIV_{}_IIV_{}', param_names=['CL', 'V'])
    assert rv1.sympy_rv.pspace.distribution.sigma == sympy.Matrix(
        [
            [symbol('OMEGA(1,1)'), symbol('IIV_CL_IIV_V')],
            [symbol('IIV_CL_IIV_V'), symbol('OMEGA(2,2)')],
        ]
    )
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2, rv3])
    rvs.join(['ETA(2)', 'ETA(3)'])
    assert rv2.sympy_rv.pspace.distribution.sigma == sympy.Matrix(
        [[symbol('OMEGA(2,2)'), 0], [0, symbol('OMEGA(3,3)')]]
    )

    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rv4, rv5 = RandomVariable.joint_normal(
        ['ETA(4)', 'ETA(5)'],
        'iiv',
        [0, 0],
        [
            [symbol('OMEGA(4,4)'), symbol('OMEGA(5,4)')],
            [symbol('OMEGA(5,4)'), symbol('OMEGA(5,5)')],
        ],
    )
    rv6 = RandomVariable.normal('EPS(1)', 'ruv', 0, symbol('SIGMA(1,1)'))
    rvs = RandomVariables([rv1, rv2, rv3, rv4, rv5, rv6])
    rvs_copy = rvs.copy()
    rvs.join(['ETA(1)', 'ETA(2)'])

    rvs = rvs_copy
    rvs.join(['ETA(1)', 'ETA(4)'])
    assert rvs['ETA(4)'].joint_names == ['ETA(1)', 'ETA(4)']
    assert rvs['ETA(5)'].joint_names == []


def test_sub():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rv3 = RandomVariable.normal('ETA(3)', 'iiv', 0, symbol('OMEGA(3,3)'))
    rvs = RandomVariables([rv1, rv2])
    rvs2 = RandomVariables([rv1, rv3])
    rvs3 = rvs - rvs2
    assert rvs3.names == ['ETA(2)']


def test_parameters_sdcorr():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable.normal('ETA(2)', 'iiv', 0, symbol('OMEGA(2,2)'))
    rvs = RandomVariables([rv1, rv2])
    params = rvs.parameters_sdcorr({'OMEGA(1,1)': 4})
    assert params == {'OMEGA(1,1)': 2}
    params = rvs.parameters_sdcorr({'OMEGA(1,1)': 4, 'OMEGA(2,2)': 16})
    assert params == {'OMEGA(1,1)': 2, 'OMEGA(2,2)': 4}

    rv1, rv2 = RandomVariable.joint_normal(
        ['ETA(1)', 'ETA(2)'],
        'iiv',
        [0, 0],
        [[symbol('x'), symbol('y')], [symbol('y'), symbol('z')]],
    )
    rvs = RandomVariables([rv1, rv2])
    params = rvs.parameters_sdcorr({'x': 4, 'y': 0.5, 'z': 16, 'k': 23})
    assert params == {'x': 2.0, 'y': 0.0625, 'z': 4.0, 'k': 23}


def test_variability_hierarchy():
    lev = VariabilityHierarchy()
    lev.add_variability_level('IIV', 0, 'ID')
    assert lev.get_name(0) == 'IIV'
    with pytest.raises(KeyError):
        lev.get_name(1)
    with pytest.raises(ValueError):
        lev.add_variability_level('IOV', 2, 'OCC')
    lev.add_variability_level('CENTER', -1, 'CENTER')
    assert len(lev) == 2
    lev.add_lower_level('PLANET', 'PLANET')
    assert len(lev) == 3
    lev.set_variability_level(-1, 'COHORT', 'X2')
    assert lev.get_name(-1) == 'COHORT'
    with pytest.raises(KeyError):
        lev.set_variability_level(2, 'X', 'Y')
    lev.remove_variability_level(-2)
    assert len(lev) == 2
    with pytest.raises(KeyError):
        lev.remove_variability_level(2)
    with pytest.raises(ValueError):
        lev.remove_variability_level('IIV')
    lev = VariabilityHierarchy()
    lev.add_variability_level('IIV', 0, 'ID')
    lev.add_higher_level('IOV', 'OCC')
    lev.add_higher_level('IOVIOV', 'SUBOCC')
    lev.add_lower_level('PLANET', 'PLANET')
    lev.add_lower_level('GALAXY', 'GALAXY')
    lev.remove_variability_level('IOV')
    assert len(lev) == 4
    lev.remove_variability_level('PLANET')
    assert len(lev) == 3


def test_covariance_matrix():
    rv1 = RandomVariable.normal('ETA(1)', 'iiv', 0, symbol('OMEGA(1,1)'))
    rv2 = RandomVariable('X', 'iiv', sympy.stats.Exponential('X', 23))
    rvs = RandomVariables([rv1, rv2])
    with pytest.raises(ValueError):
        rvs.covariance_matrix
    rvs = RandomVariables([])
    assert len(rvs.covariance_matrix) == 0



from .tool import (
    Modelfit,
    create_multiple_fit_workflow,
    create_single_fit_workflow,
    create_workflow,
)

__all__ = ['Modelfit', create_workflow, create_single_fit_workflow, create_multiple_fit_workflow]



# Functional interface to extract dataset information


def get_number_of_individuals(model):
    """Retrieve the number of individuals in the model dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of individuals in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_individuals, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_individuals(model)
    59

    Notes
    -----
    For NONMEM models this is the number of individuals of the active dataset, i.e. after filtering
    of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_observations : Get the number of observations in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    return model.dataset.pharmpy.ninds


def get_number_of_observations(model):
    """Retrieve the total number of observations in the model dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    int
        Number of observations in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_observations, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_observations(model)
    155

    Notes
    -----
    For NONMEM models this is the number of observations of the active dataset, i.e. after filtering
    of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_individuals : Get the number of individuals in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    return model.dataset.pharmpy.nobs


def get_number_of_observations_per_individual(model):
    """Number of observations for each individual

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        Number of observations in the model dataset

    Examples
    --------
    >>> from pharmpy.modeling import get_number_of_observations_per_individual, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_number_of_observations_per_individual(model)
    ID
    1     2
    2     3
    3     3
    4     3
    5     3
    6     3
    7     3
    8     3
    9     4
    10    3
    11    1
    12    3
    13    2
    14    4
    15    2
    16    3
    17    3
    18    4
    19    3
    20    3
    21    3
    22    2
    23    3
    24    3
    25    6
    26    2
    27    2
    28    1
    29    1
    30    2
    31    1
    32    3
    33    2
    34    2
    35    2
    36    3
    37    2
    38    4
    39    3
    40    2
    41    3
    42    2
    43    1
    44    3
    45    3
    46    1
    47    1
    48    5
    49    3
    50    4
    51    3
    52    3
    53    2
    54    4
    55    1
    56    1
    57    2
    58    3
    59    3
    Name: observation_count, dtype: int64

    Notes
    -----
    For NONMEM models this is the individuals and number of observations of the active dataset, i.e.
    after filtering of IGNORE and ACCEPT and removal of individuals with no observations.

    See also
    --------
    get_number_of_individuals : Get the number of individuals in a dataset
    get_number_of_observations_per_individual : Get the number of observations per individual in a
        dataset

    """
    return model.dataset.pharmpy.nobsi


def get_observations(model):
    """Get observations from dataset

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    pd.Series
        Observations indexed over ID and TIME

    Examples
    --------
    >>> from pharmpy.modeling import get_observations, load_example_model
    >>> model = load_example_model("pheno")
    >>> get_observations(model)
        ID  TIME
    1   2.0      17.3
        112.5    31.0
    2   2.0       9.7
        63.5     24.6
        135.5    33.0
                 ...
    58  47.5     27.9
        131.8    31.0
    59  1.8      22.6
        73.8     34.3
        146.8    40.2
    Name: DV, Length: 155, dtype: float64

    See also
    --------
    get_number_of_observations
    get_number_of_observations_per_individual

    """
    return model.dataset.pharmpy.observations



import enum
from pathlib import Path

import numpy as np
import pandas as pd

import pharmpy.data


class DatasetError(Exception):
    pass


class DatasetWarning(Warning):
    pass


class ColumnType(enum.Enum):
    """The type of the data in a column"""

    UNKNOWN = enum.auto()
    ID = enum.auto()
    IDV = enum.auto()
    DV = enum.auto()
    COVARIATE = enum.auto()
    DOSE = enum.auto()
    EVENT = enum.auto()

    def __repr__(self):
        return f'{self.__class__.__name__}.{self.name}'

    @property
    def max_one(self):
        """Can this ColumnType only be assigned to at most one column?"""
        return self in (ColumnType.ID, ColumnType.IDV, ColumnType.DOSE, ColumnType.EVENT)


class PharmDataFrame(pd.DataFrame):
    """A DataFrame with additional metadata.

    Each column can have a ColumnType. The default ColumnType is UNKNOWN.

    ============  =============
    ColumnType    Description
    ============  =============
    ID            Individual identifier. Max one per DataFrame. All values have to be unique
    IDV           Independent variable. Max one per DataFrame.
    DV            Dependent variable
    COVARIATE     Covariate
    DOSE          Dose amount
    EVENT         0 = observation
    UNKOWN        Unkown type. This will be the default for columns that hasn't been assigned a type
    ============  =============

    """

    _metadata = ['_column_types', 'name']

    @property
    def _constructor(self):
        return PharmDataFrame

    @property
    def _constructor_sliced(self):
        return pd.Series

    def __deepcopy__(self, memo):
        return self.copy()

    def copy(self, *kwargs):
        """ """
        # FIXME: Set empty docstring to avoid getting documentation from base class
        #        would like sphinx to do this so that in object docstring is kept.
        new_df = super().copy(*kwargs)
        try:
            new_df._column_types = self._column_types.copy()
        except AttributeError:
            pass
        return new_df

    def to_json(self, **kwargs):
        """ """
        # FIXME: Same docstring issue as for copy
        # FIXME: Directly using to_json on PharmDataFrame doesn't work
        return pd.DataFrame(self).to_json(**kwargs)


class ColumnTypeIndexer:
    """Indexing a PharmDataFrame to get or set column types
    An instance of ColumnTypeIndexer can be retrieved by df.pharmpy.column_type
    """

    def __init__(self, df):
        self._obj = df

    def _set_one_column_type(self, label, tp):
        if label not in self._obj.columns:
            raise KeyError(str(label))
        if (
            tp.max_one
            and hasattr(self._obj, '_column_types')
            and tp in self._obj._column_types.values()
        ):
            raise KeyError(f'Only one column of type {tp} is allowed in a PharmDataFrame.')
        try:
            self._obj._column_types[label] = tp
        except AttributeError:
            self._obj._column_types = {}
            self._obj._column_types[label] = tp

    def __setitem__(self, ind, tp):
        if isinstance(ind, str):
            self._set_one_column_type(ind, tp)
        else:
            if hasattr(tp, '__len__') and not len(tp) == 1:
                if len(ind) == len(tp):
                    for label, one_tp in zip(ind, tp):
                        self._set_one_column_type(label, one_tp)
                else:
                    raise ValueError(f'Cannot set {len(ind)} columns using {len(tp)} column types')
            else:
                # Broadcasting of tp
                try:
                    len(ind)
                except TypeError:
                    ind = [ind]
                for label in ind:
                    self._set_one_column_type(label, tp)

    def _get_one_column_type(self, label):
        """Get the column type of one column"""
        if label in self._obj.columns:
            try:
                d = self._obj._column_types
            except AttributeError:
                return ColumnType.UNKNOWN
            try:
                return d[label]
            except KeyError:
                return ColumnType.UNKNOWN
        else:
            raise KeyError(str(label))

    def __getitem__(self, ind):
        if isinstance(ind, str):
            return self._get_one_column_type(ind)
        else:
            try:
                return [self._get_one_column_type(label) for label in ind]
            except TypeError:
                return self._get_one_column_type(ind)


class LabelsByTypeIndexer:
    """Indexing a PharmDataFrame to get labels from ColumnTypes"""

    def __init__(self, acc):
        self._acc = acc

    def _get_one_label(self, tp):
        labels = self._get_many_labels(tp)
        if len(labels) == 0:
            raise KeyError(str(tp))
        elif len(labels) > 1:
            raise KeyError('Did not expect more than one ' + str(tp))
        else:
            return labels

    def _get_many_labels(self, column_type):
        """Will raise if no columns of the type exists
        Always returns a list of labels
        """
        return [
            label for label in self._acc._obj.columns if self._acc.column_type[label] == column_type
        ]

    def __getitem__(self, tp):
        try:
            len(tp)
            labels = []
            for t in tp:
                if t.max_one:
                    labels.extend(self._get_one_label(t))
                else:
                    labels.extend(self._get_many_labels(t))
            return labels
        except TypeError:
            if tp.max_one:
                return self._get_one_label(tp)
            else:
                return self._get_many_labels(tp)


@pd.api.extensions.register_dataframe_accessor('pharmpy')
class DataFrameAccessor:
    def __init__(self, obj):
        self._obj = obj

    @property
    def column_type(self):
        return ColumnTypeIndexer(self._obj)

    @property
    def labels_by_type(self):
        return LabelsByTypeIndexer(self)

    @property
    def id_label(self):
        """Return the label of the id column"""
        return self.labels_by_type[ColumnType.ID][0]

    @property
    def idv_label(self):
        """Return the label of the idv column"""
        return self.labels_by_type[ColumnType.IDV][0]

    @property
    def dv_label(self):
        """Return the label of the dv column"""
        return self.labels_by_type[ColumnType.DV][0]

    @property
    def ids(self):
        """Return the ids in the dataset"""
        return self._obj[self.id_label].unique()

    @property
    def time_varying_covariates(self):
        """Return a list of labels for all time varying covariates"""
        cov_labels = self.labels_by_type[ColumnType.COVARIATE]
        if len(cov_labels) == 0:
            return []
        else:
            time_var = self._obj.groupby(by=self.id_label)[cov_labels].nunique().gt(1).any()
            return list(time_var.index[time_var])

    @property
    def baselines(self):
        """Baselines for each id.
        Baseline is taken to be the first row even if that has a missing value.
        """
        idlab = self.id_label
        return self._obj.groupby(idlab).nth(0)

    @property
    def covariate_baselines(self):
        """Return a dataframe with baselines of all covariates for each id.
        Baseline is taken to be the first row even if that has a missing value.
        """
        covariates = self.labels_by_type[ColumnType.COVARIATE]
        idlab = self.id_label
        df = self._obj[covariates + [idlab]]
        df.set_index(idlab, inplace=True)
        return df.groupby(idlab).nth(0)

    @property
    def observations(self):
        """Return a series with observations. Indexed with ID and TIME"""
        try:
            label = self.labels_by_type[ColumnType.EVENT]
        except KeyError:
            try:
                label = self.labels_by_type[ColumnType.DOSE]
            except KeyError:
                raise DatasetError('Could not identify observation rows in dataset')

        label = label[0]
        idcol = self.id_label
        idvcol = self.idv_label
        df = self._obj.query(f'{label} == 0')

        if df.empty:
            df = self._obj.astype({label: 'float'})
            df = df.query(f'{label} == 0')

        df = df[[idcol, idvcol, self.dv_label]]
        try:
            # FIXME: This shouldn't be needed
            df = df.astype({idvcol: np.float64})
        except ValueError:
            # TIME could not be converted to float (e.g. 10:15)
            pass
        df.set_index([idcol, idvcol], inplace=True)
        return df.squeeze()

    @property
    def ninds(self):
        return len(self.ids)

    @property
    def nobs(self):
        return len(self.observations)

    @property
    def nobsi(self):
        ser = self.observations.groupby(self.id_label).count()
        ser.name = "observation_count"
        return ser

    @property
    def doses(self):
        """Return a series with all doses. Indexed with ID and TIME"""
        try:
            label = self.labels_by_type[ColumnType.DOSE]
        except KeyError:
            raise DatasetError('Could not identify dosing rows in dataset')

        label = label[0]
        idcol = self.id_label
        idvcol = self.idv_label
        df = self._obj.query(f'{label} != 0')
        df = df[[idcol, idvcol, label]]
        df.set_index([idcol, idvcol], inplace=True)
        return df.squeeze()

    def add_doseid(self):
        """Add a column DOSEID with id of each dose period starting from 1"""
        try:
            dose = self.labels_by_type[ColumnType.DOSE]
        except KeyError:
            raise DatasetError('Could not identify dosing rows in dataset')
        df = self._obj
        df['DOSEID'] = df[dose]
        df.loc[df['DOSEID'] > 0, 'DOSEID'] = 1
        df['DOSEID'] = df['DOSEID'].astype(int)
        df['DOSEID'] = df.groupby(self.id_label)['DOSEID'].cumsum()

    def add_time_after_dose(self):
        """Calculate and add a TAD column to the dataset"""
        # FIXME: Should not rely on name here. Use coltypes for TAD and DOSEID
        # FIXME: TIME is converted to float. Should be handled when reading in dataset
        df = self._obj
        if 'DOSEID' in df.columns:
            had_doseid = True
        else:
            self.add_doseid()
            had_doseid = False
        idv = self.idv_label
        idlab = self.id_label
        df[idv] = df[idv].astype(np.float64)
        df['TAD'] = df.groupby([idlab, 'DOSEID'])[idv].diff().fillna(0)
        df['TAD'] = df.groupby([idlab, 'DOSEID'])['TAD'].cumsum()
        if not had_doseid:
            df.drop('DOSEID', axis=1, inplace=True)

    def concentration_parameters(self):
        """Create a dataframe with concentration parameters"""
        df = self._obj.copy()
        df.pharmpy.add_doseid()
        df.pharmpy.add_time_after_dose()
        idlab = self.id_label
        dv = self.dv_label
        noobs = df.groupby([idlab, 'DOSEID']).size() == 1
        idx = df.groupby([idlab, 'DOSEID'])[dv].idxmax()
        params = df.loc[idx].set_index([idlab, 'DOSEID'])
        params = params[[dv, 'TAD']]
        params.rename(columns={dv: 'Cmax', 'TAD': 'Tmax'}, inplace=True)
        params.loc[noobs] = np.nan

        grpind = df.groupby(['ID', 'DOSEID']).indices
        keep = []
        for ind, rows in grpind.items():
            index = idx.loc[ind]
            p = params.loc[ind]
            if not np.isnan(p['Tmax']):
                keep += [row for row in rows if row > index]
        minidx = df.iloc[keep].groupby([idlab, 'DOSEID'])[dv].idxmin()
        params2 = df.loc[minidx].set_index([idlab, 'DOSEID'])
        params2 = params2[[dv, 'TAD']]
        params2.rename(columns={dv: 'Cmin', 'TAD': 'Tmin'}, inplace=True)
        res = params.join(params2)
        return res

    def generate_path(self, path=None, force=False):
        """Generate the path of dataframe if written.
        If no path is supplied or does not contain a filename a name is created
        from the name property of the PharmDataFrame.
        Will not overwrite unless forced.
        """
        if path is None:
            path = Path("")
        else:
            path = Path(path)
        if not path or path.is_dir():
            try:
                filename = f'{self._obj.name}.csv'
            except AttributeError:
                raise ValueError(
                    'Cannot name data file as no path argument was supplied and '
                    'the DataFrame has no name property.'
                )
            path /= filename
        if not force and path.exists():
            raise FileExistsError(f'File at {path} already exists.')
        return path

    def write_csv(self, path=None, force=False):
        """Write PharmDataFrame to a csv file
        return path for the written file
        """
        path = self.generate_path(path, force)
        self._obj.to_csv(path, na_rep=pharmpy.data.conf.na_rep, index=False)
        return path



import pytest

import pharmpy.plugins.nonmem.nmtran_parser
import pharmpy.plugins.nonmem.records


@pytest.fixture
def records():
    return pharmpy.plugins.nonmem.records


@pytest.fixture
def parser():
    return pharmpy.plugins.nonmem.nmtran_parser.NMTranParser()



import pharmpy.results
import pharmpy.tools.iiv.algorithms as algorithms
from pharmpy.tools.modelfit import create_single_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(algorithm, rankfunc='ofv', cutoff=None, model=None):
    algorithm_func = getattr(algorithms, algorithm)

    wf = Workflow()
    wf.name = "iiv"

    if model is not None:
        start_task = Task('start_iiv', start, model)
    else:
        start_task = Task('start_iiv', start)

    wf.add_task(start_task)

    if model and not model.modelfit_results:
        wf_fit = create_single_fit_workflow()
        wf.insert_workflow(wf_fit, predecessors=start_task)
        start_model_task = wf_fit.output_tasks
    else:
        start_model_task = [start_task]

    wf_method = algorithm_func(model)
    wf.insert_workflow(wf_method)

    task_result = Task(
        'results',
        post_process_results,
        rankfunc,
        cutoff,
    )

    wf.add_task(task_result, predecessors=start_model_task + wf.output_tasks)

    return wf


def start(model):
    return model


def post_process_results(rankfunc, cutoff, *models):
    res_models = []
    model_features = dict()

    for model in models:
        if not model.name.startswith('iiv_candidate'):
            start_model = model
        else:
            res_models.append(model)
            model_features[model.name] = _get_iiv_block(model.random_variables)

    df = pharmpy.tools.common.create_summary(
        res_models, start_model, rankfunc, cutoff, model_features
    )

    best_model_name = df['rank'].idxmin()
    try:
        best_model = [model for model in res_models if model.name == best_model_name][0]
    except IndexError:
        best_model = None

    res = IIVResults(summary=df, best_model=best_model, start_model=start_model, models=res_models)

    return res


def _get_iiv_block(rvs):
    return [eta.name for eta in rvs.iiv if len(eta.joint_names) > 0]


class IIVResults(pharmpy.results.Results):
    def __init__(self, summary=None, best_model=None, start_model=None, models=None):
        self.summary = summary
        self.best_model = best_model
        self.start_model = start_model
        self.models = models



import pharmpy.config as config

from .local import LocalDispatcher


class DispatcherConfiguration(config.Configuration):
    module = 'pharmpy.workflows.dispatchers'
    dask_dispatcher = config.ConfigItem(
        None,
        'Which type of dask scheduler to use (supports threaded and distributed).',
        str,
    )


conf = DispatcherConfiguration()


__all__ = ['LocalDispatcher']



import importlib

import pharmpy.model
import pharmpy.results
import pharmpy.tools.common
import pharmpy.tools.modelfit
from pharmpy.workflows import execute_workflow

from .common import read_model_from_database


def fit(models):
    """Fit models.

    Parameters
    ----------
    models : list
        List of models or one single model

    Return
    ------
    Model
        Reference to same model

    Examples
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> fit(model)      # doctest: +SKIP

    See also
    --------
    run_tool

    """
    if isinstance(models, pharmpy.model.Model):
        models = [models]
        single = True
    else:
        single = False
    kept = []
    # Do not fit model if already fit
    for model in models:
        try:
            db_model = read_model_from_database(model.name, database=model.database)
        except KeyError:
            db_model = None
        if db_model and db_model.modelfit_results is not None and db_model == model:
            model.modelfit_results = db_model.modelfit_results
        else:
            kept.append(model)
    if kept:
        run_tool('modelfit', kept)
    if single:
        return models[0]
    else:
        return models


def create_results(path, **kwargs):
    """Create/recalculate results object given path to run directory

    Parameters
    ----------
    path : str, Path
        Path to run directory
    kwargs
        Arguments to pass to tool specific create results function

    Returns
    -------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> res = create_results("frem_dir1")   # doctest: +SKIP

    See also
    --------
    read_results

    """
    res = pharmpy.tools.common.create_results(path, **kwargs)
    return res


def read_results(path):
    """Read results object from file

    Parameters
    ----------
    path : str, Path
        Path to results file

    Return
    ------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> res = read_resuts("results.json")     # doctest: +SKIP

    See also
    --------
    create_results

    """
    res = pharmpy.results.read_results(path)
    return res


def run_tool(name, *args, **kwargs):
    """Run tool workflow

    Parameters
    ----------
    name : str
        Name of tool to run
    args
        Arguments to pass to tool
    kwargs
        Arguments to pass to tool

    Return
    ------
    Results
        Results object for tool

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> res = run_tool("resmod", model)   # doctest: +SKIP

    """
    tool = importlib.import_module(f'pharmpy.tools.{name}')
    wf = tool.create_workflow(*args, **kwargs)
    res = execute_workflow(wf)
    return res



from io import StringIO

import pharmpy
from pharmpy.data import ColumnType


def test_data_read(pheno):
    df = pheno.dataset
    # FIXME: should have numeric TIME
    assert list(df.iloc[1]) == [1.0, '2.0', 0.0, 1.4, 7.0, 17.3, 0.0, 0.0]
    assert list(df.columns) == ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV', 'FA1', 'FA2']


def test_read_raw_dataset(pheno):
    df = pheno.read_raw_dataset()
    assert list(df.iloc[0]) == ['1', '0.', '25.0', '1.4', '7', '0', '1', '1']
    assert list(df.columns) == ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV', 'FA1', 'FA2']
    assert df.pharmpy.column_type['ID'] == ColumnType.ID


def test_ignore_with_synonym(pheno_data):
    model = pharmpy.Model(
        StringIO(
            f"$PROBLEM dfs\n$INPUT ID TIME AMT WT APGR DV=CONC FA1 FA2\n"
            f"$DATA {pheno_data} IGNORE=@ IGNORE=(CONC.EQN.0)"
        )
    )
    df = model.dataset
    assert len(df) == 155
    model = pharmpy.Model(
        StringIO(
            f"$PROBLEM dfs\n$INPUT ID TIME AMT WT APGR DV=CONC FA1 FA2\n"
            f"$DATA {pheno_data} IGNORE=@ IGNORE=(DV.EQN.0)"
        )
    )
    df = model.dataset
    assert len(df) == 155


def test_idv_with_synonym(pheno_data):
    model = pharmpy.Model(
        StringIO(
            f"$PROBLEM dfs\n$INPUT ID TIME=TAD AMT WT APGR DV FA1 FA2\n"
            f"$DATA {pheno_data} IGNORE=@"
        )
    )
    df = model.dataset

    assert df.pharmpy.column_type['TAD'] == ColumnType.IDV
